from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Dict

import pandas as pd

if TYPE_CHECKING:
    from ..core.graph import TypeGraph


class Base:

    def __init__(
            self,
            graph: TypeGraph,
    ):
        self.graph = graph

    def get_data_frame_plot(
            self,
            *args,
            **kwargs,
    ) -> pd.DataFrame:
        df = self.graph.nodes.update(data=self.graph.nodes.get_positions()).data
        return df

    def plot(self, **kwargs):
        ...


class Plotly(Base):

    def plot(
            self,
            df: Optional[pd.DataFrame] = None,
            x: Optional[str] = None,
            y: Optional[str] = None,
            hover_data: Optional[Dict[str]] = None,
            size_default: Optional[int] = None,
            **kwargs,
    ):
        import plotly.express as px

        if df is None:
            df = self.get_data_frame_plot()

        if x is None:
            x = 'position_dim_0'

        if y is None:
            y = 'position_dim_1'

        if hover_data is None:
            hover_data = dict()
        hover_data['node'] = True
        hover_data['position_dim_0'] = False
        hover_data['position_dim_1'] = False

        if size_default is None:
            size_default = 50

        if df.index.name == 'node':
            df = df.reset_index()

        fig = (
            px.scatter(
                df,
                x=x,
                y=y,
                hover_data=hover_data,
                **kwargs,
            )
            .update_xaxes(visible=False, showgrid=False)
            .update_yaxes(visible=False, showgrid=False)
            .for_each_trace(
                lambda trace: trace.update(marker_size=size_default) if trace['marker_size'] is None else (),
            )
        )

        # for edge:
        # https://stackoverflow.com/questions/51410283/how-to-efficiently-create-interactive-directed-network-graphs-with-arrows-on-p

        return fig



from __future__ import annotations
from typing import Optional, Union, TypeVar, Callable, Sequence, Dict

from pandas import DataFrame, MultiIndex, concat as pandas_concat

from .base import (
    TypeDataInput,
    TypeMapper,
)
from .nodes import (
    Nodes,
    TypeNodesInput
)
from .edges import (
    Edges,
    UndirectedEdges,
    TypeEdgesInput,
)
from ..plot.engine import Plotly as PlotlyEngine

__all__ = [
    'Graph',
    'UndirectedGraph',
    'TypeGraph',
]

TypeGraph = TypeVar('TypeGraph', bound='Graph')


class Graph:

    _type_edges = Edges
    _columns_degree_keep = ['degree_in', 'degree_out', 'degree']
    _columns_neighbors_keep = ['neighbors_in', 'neighbors_out', 'neighbors']

    def __init__(
            self,
            nodes: Optional[TypeNodesInput] = None,
            edges: Optional[TypeEdgesInput] = None,
            order: Optional[int] = None,
            nodes_data: Optional[TypeDataInput] = None,
            edges_data: Optional[TypeDataInput] = None,
            **kwargs,
    ):
        self._nodes = Nodes(index=nodes, data=nodes_data, order=order, **kwargs)
        self._edges = self._type_edges(index=edges, data=edges_data, **kwargs)
        self._nodes = self._update_nodes_from_edges()
        self._engine_plot = PlotlyEngine(graph=self)


    @classmethod
    def path(
            cls,
            order: Optional[int] = None,
            nodes: Optional[Nodes] = None,
    ) -> TypeGraph:

        if order is None:
            order = len(nodes)

        graph = cls(order=order, edges=MultiIndex.from_arrays([range(order - 1), range(1, order)]))

        if nodes is not None:
            graph = graph.map_nodes(mapper=dict(zip(range(order), Nodes(index=nodes))))

        return graph

    @classmethod
    def cycle(
            cls,
            order: Optional[int] = None,
            nodes: Optional[Nodes] = None,
    ) -> TypeGraph:

        if order is None:
            order = len(nodes)

        if order < 3:
            raise ValueError("Inputs 'order' or length of 'nodes' has to be >= 3 for cycle")

        nodes_left = range(order)
        nodes_right = list(range(1, order))
        nodes_right.append(0)
        edges = MultiIndex.from_arrays([nodes_left, nodes_right])

        graph = cls(order=order, edges=edges)

        if nodes is not None:
            graph = graph.map_nodes(mapper=dict(zip(range(order), Nodes(index=nodes))))

        return graph

    @classmethod
    def star(
            cls,
            order: Optional[int] = None,
            nodes: Optional[Nodes] = None,
    ) -> TypeGraph:

        if order is None:
            order = len(nodes)

        graph = cls(order=order, edges=MultiIndex.from_arrays([[0] * (order - 1), range(1, order)]))

        if nodes is not None:
            graph = graph.map_nodes(mapper=dict(zip(range(order), Nodes(index=nodes))))

        return graph

    @classmethod
    def _get_unique_edges(
            cls,
            edges: MultiIndex,
    ) -> MultiIndex:
        return edges

    @classmethod
    def complete(
            cls,
            order: Optional[int] = None,
            nodes: Optional[Nodes] = None,
    ) -> TypeGraph:

        if order is None:
            order = len(nodes)

        edges = MultiIndex.from_product([range(order)] * 2)
        edges = edges[edges.get_level_values(0) != edges.get_level_values(1)]
        edges = cls._get_unique_edges(edges=edges)

        graph = cls(order=order, edges=edges)

        if nodes is not None:
            graph = graph.map_nodes(mapper=dict(zip(range(order), Nodes(index=nodes))))

        return graph

    @property
    def nodes(self) -> Nodes:
        return self._nodes

    @nodes.setter
    def nodes(
            self,
            value: TypeNodesInput,
    ):
        self._nodes = Nodes(index=value)
        self._update_edges_from_nodes()

    @property
    def edges(self) -> Edges:
        return self._edges

    @edges.setter
    def edges(
            self,
            value: TypeEdgesInput,
    ):
        self._edges = self._type_edges(index=value)
        self._update_nodes_from_edges()

    def assign(
            self,
            nodes: Optional[Union[Nodes, Callable]] = None,
            edges: Optional[Union[Edges, Callable]] = None,
    ) -> TypeGraph:
        if nodes is not None:
            if isinstance(nodes, Callable):
                nodes = nodes(self)
            self.nodes = nodes

        if edges is not None:
            if isinstance(edges, Callable):
                edges = edges(self)
            self.edges = edges

        return self

    def _update_nodes_from_edges(self) -> Nodes:
        return self.nodes.update(index=self.edges.nodes)

    def _update_edges_from_nodes(self) -> Edges:
        return self.edges.keep_nodes(nodes=self.nodes)

    def update_nodes(
            self,
            nodes: Optional[TypeNodesInput] = None,
            data: Optional[TypeDataInput] = None,
            **kwargs,
    ) -> TypeGraph:
        return self.assign(nodes=self.nodes.update(index=nodes, data=data, **kwargs))

    def update_edges(
            self,
            edges: Optional[TypeEdgesInput] = None,
            data: Optional[TypeDataInput] = None,
            **kwargs,
    ) -> TypeGraph:
        return self.assign(edges=self.edges.update(index=edges, data=data, **kwargs))

    def drop_nodes(
            self,
            nodes: Optional[TypeNodesInput] = None,
            **kwargs,
    ) -> TypeGraph:
        self.assign(nodes=self.nodes.drop(index=nodes, **kwargs))
        return self

    def drop_edges(
            self,
            edges: Optional[TypeEdgesInput] = None,
            **kwargs,
    ) -> TypeGraph:
        return self.assign(edges=self.edges.drop(index=edges, **kwargs))

    @property
    def order(self) -> int:
        return len(self.nodes)

    @property
    def size(self) -> int:
        return len(self.edges)

    def __repr__(self):
        return f'{self.__class__.__name__}(nodes={self.nodes}, edges={self.edges})'

    def map_nodes(
            self,
            mapper: Optional[TypeMapper] = None,
            **kwargs
    ) -> TypeGraph:
        self._nodes = self.nodes.map(mapper=mapper, **kwargs)
        self._edges = self.edges.map(mapper=mapper, **kwargs)
        return self

    def _edges_stack_value_counts(
            self,
            column_name: str,
            columns_stack: Optional[Sequence[str]] = None,
    ) -> DataFrame:

        if columns_stack is None:
            columns_stack = self.edges.names_index

        df = (
            self.edges
            .data_internal
            [columns_stack]
            .stack()
            .value_counts()
            .reindex(self.nodes.data.index)
            .fillna(0)
            .rename(column_name)
            .to_frame()
        )

        return df

    def _info_columns_degree(self) -> Dict[str, Dict[str, Sequence[str]]]:
        info = {
            'degree': dict(),
            'degree_in': dict(columns_stack=[self.edges.names_index[1]]),
            'degree_out': dict(columns_stack=[self.edges.names_index[0]]),
        }
        return info

    @property
    def degrees(self) -> DataFrame:

        info_columns_degree = self._info_columns_degree()

        df = Nodes(data=pandas_concat(
            [
                self._edges_stack_value_counts(column_name=column, **info_columns_degree[column])
                for column in self._columns_degree_keep
            ],
            axis=1,
        )).data

        return df

    @property
    def neighbors(self) -> DataFrame:

        column_neighbors_in = self.edges.names_index[0].replace(self.nodes.names_index[0], 'neighbors')
        column_neighbors_out = self.edges.names_index[1].replace(self.nodes.names_index[0], 'neighbors')

        neighbors_in = (
            self.edges
            .data_internal
            .groupby(self.edges.names_index[1])
            [self.edges.names_index[0]]
            .apply(list)
            .rename_axis(index=self.nodes.names_index[0])
            .rename(column_neighbors_in)
        )

        neighbors_out = (
            self.edges
            .data_internal
            .groupby(self.edges.names_index[0])
            [self.edges.names_index[1]]
            .apply(list)
            .rename_axis(index=self.nodes.names_index[0])
            .rename(column_neighbors_out)
        )

        df = Nodes(data=(
            pandas_concat([neighbors_in, neighbors_out], axis=1)
            .reindex(self.nodes.data.index)
            .applymap(lambda x: x if isinstance(x, list) else [])
            .assign(neighbors=lambda x: (x[column_neighbors_in] + x[column_neighbors_out]).apply(frozenset).apply(list))
            [self._columns_neighbors_keep]
        )).data

        return df

    def plot(
            self,
            **kwargs,
    ):
        return self._engine_plot.plot(**kwargs)


class UndirectedGraph(Graph):

    _type_edges = UndirectedEdges
    _columns_degree_keep = ['degree']
    _columns_neighbors_keep = ['neighbors']

    @classmethod
    def _get_unique_edges(
            cls,
            edges: MultiIndex,
    ) -> MultiIndex:
        return edges.map(frozenset).drop_duplicates().map(tuple)



from __future__ import annotations
from typing import Optional, Union, TypeVar

from .base import (
    ContainerDataIndexedMulti,
    ContainerDataIndexedMultiSets,
    TypeIndexMultiInput,
)
from .nodes import (
    Nodes,
    TypeNodesInput
)

__all__ = [
    'Edges',
    'UndirectedEdges',
    'TypeEdgesInput',
]

TypeEdges = TypeVar('TypeEdges', bound='Edges')


class Edges(ContainerDataIndexedMulti):

    _names_index = ['node_in', 'node_out']
    _name_index_internal = 'edge_internal'

    @property
    def nodes(self) -> Nodes:
        return Nodes(index=self.data_internal[self._names_index].stack().drop_duplicates())

    def keep_nodes(
            self,
            nodes: Optional[TypeNodesInput] = None,
    ) -> TypeEdges:
        if nodes is not None:
            nodes = Nodes(index=nodes)
            self.data = self.data_internal[
                lambda x: x[self._names_index[0]].isin(nodes) & x[self._names_index[1]].isin(nodes)
            ]
        return self


class UndirectedEdges(ContainerDataIndexedMultiSets, Edges):
    _names_index = ['node_0', 'node_1']
    _name_index_internal = 'edge_internal'


TypeEdgesInput = Union[Edges, TypeIndexMultiInput]



import pytest

import tahini.testing
import tahini.core.base
import tahini.core


@pytest.mark.parametrize('args, kwargs, type_error, message_error', [
    # mismatch in types
    (
        [],
        dict(left=tahini.core.base.ContainerDataIndexed(), right=tahini.core.base.ContainerDataIndexedMulti()),
        AssertionError,
        """Types are different

[left]:  ContainerDataIndexed
[right]: ContainerDataIndexedMulti""",
    ),
    (
        [],
        dict(left=tahini.core.base.ContainerDataIndexedMulti(), right=tahini.core.base.ContainerDataIndexedMultiSets()),
        AssertionError,
        """Types are different

[left]:  ContainerDataIndexedMulti
[right]: ContainerDataIndexedMultiSets""",
    ),
    # different length
    (
        [],
        dict(left=tahini.core.base.ContainerDataIndexed(), right=tahini.core.base.ContainerDataIndexed(index=[0])),
        AssertionError,
        """ContainerDataIndexed.data_testing are different

ContainerDataIndexed.data_testing shape mismatch
[left]:  (0, 0)
[right]: (1, 0)""",
    ),
    # different index
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexed(index=[0]),
            right=tahini.core.base.ContainerDataIndexed(index=[1]),
        ),
        AssertionError,
        """ContainerDataIndexed.data_testing.index are different

ContainerDataIndexed.data_testing.index values are different (100.0 %)
[left]:  Int64Index([0], dtype='int64')
[right]: Int64Index([1], dtype='int64')""",
    ),
    # different index type
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexed(index=[0]),
            right=tahini.core.base.ContainerDataIndexed(index=['a']),
        ),
        AssertionError,
        """ContainerDataIndexed.data_testing.index are different

ContainerDataIndexed.data_testing.index values are different (100.0 %)
[left]:  Int64Index([0], dtype='int64')
[right]: Index(['a'], dtype='object')""",
    ),
    # different order
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
            right=tahini.core.base.ContainerDataIndexedMulti(index=[(1, 0)]),
        ),
        AssertionError,
        """ContainerDataIndexedMulti.data_testing.index are different

ContainerDataIndexedMulti.data_testing.index values are different (100.0 %)
[left]:  Index([(0, 1)], dtype='object')
[right]: Index([(1, 0)], dtype='object')""",
    ),
])
def test_assert_container_equal_error(args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        tahini.testing.testing.assert_container_equal(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('args, kwargs', [
    # empty
    ([], dict(left=tahini.core.base.ContainerDataIndexed(), right=tahini.core.base.ContainerDataIndexed())),
    ([], dict(left=tahini.core.base.ContainerDataIndexedMulti(), right=tahini.core.base.ContainerDataIndexedMulti())),
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexedMultiSets(),
            right=tahini.core.base.ContainerDataIndexedMultiSets()),
    ),
    # non empty
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexed(index=[0]),
            right=tahini.core.base.ContainerDataIndexed(index=[0])),
    ),
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexed(index=[0, 1]),
            right=tahini.core.base.ContainerDataIndexed(index=[0, 1])),
    ),
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexed(index=[0, 1]),
            right=tahini.core.base.ContainerDataIndexed(index=[1, 0])),
    ),
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
            right=tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)])),
    ),
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1), (0, 2)]),
            right=tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1), (0, 2)])),
    ),
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
            right=tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)])),
    ),
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1), (0, 2)]),
            right=tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1), (0, 2)])),
    ),
    # order does not matter for ContainerDataIndexedMultiSets
    (
        [],
        dict(
            left=tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
            right=tahini.core.base.ContainerDataIndexedMultiSets(index=[(1, 0)])),
    ),
])
def test_assert_container_equal(args, kwargs):
    tahini.testing.testing.assert_container_equal(*args, **kwargs)


@pytest.mark.parametrize('args, kwargs, type_error, message_error', [
    # different type
    (
        [],
        dict(left=tahini.core.Graph(), right=tahini.core.UndirectedGraph()),
        AssertionError,
        '''Types are different

[left]:  Edges
[right]: UndirectedEdges''',
    ),
    # different length nodes
    (
        [],
        dict(left=tahini.core.Graph(), right=tahini.core.Graph(order=1)),
        AssertionError,
        '''Graph.nodes.data_testing are different

Graph.nodes.data_testing shape mismatch
[left]:  (0, 0)
[right]: (1, 0)''',
    ),
    # different length nodes for undirected
    (
        [],
        dict(left=tahini.core.UndirectedGraph(), right=tahini.core.UndirectedGraph(order=1)),
        AssertionError,
        '''UndirectedGraph.nodes.data_testing are different

UndirectedGraph.nodes.data_testing shape mismatch
[left]:  (0, 0)
[right]: (1, 0)''',
    ),
    # different length edges
    (
        [],
        dict(left=tahini.core.Graph(nodes=[0, 1]), right=tahini.core.Graph(edges=[(0, 1)])),
        AssertionError,
        '''Graph.edges.data_testing are different

Graph.edges.data_testing shape mismatch
[left]:  (0, 0)
[right]: (1, 0)''',
    ),
    # order of each edge matters for graph
    (
        [],
        dict(
            left=tahini.core.Graph(edges=[(0, 1)]),
            right=tahini.core.Graph(edges=[(1, 0)]),
        ),
        AssertionError,
        '''Graph.edges.data_testing.index are different

Graph.edges.data_testing.index values are different (100.0 %)
[left]:  Index([(0, 1)], dtype='object')
[right]: Index([(1, 0)], dtype='object')''',
    ),
    # non default obj
    (
        [],
        dict(left=tahini.core.Graph(nodes=[0, 1]), right=tahini.core.Graph(edges=[(0, 1)]), obj='ChildGraph'),
        AssertionError,
        '''ChildGraph are different

ChildGraph shape mismatch
[left]:  (0, 0)
[right]: (1, 0)''',
    ),
])
def test_assert_graph_equal_error(args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        tahini.testing.testing.assert_graph_equal(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('args, kwargs', [
    # empty
    ([], dict(left=tahini.core.Graph(), right=tahini.core.Graph())),
    ([], dict(left=tahini.core.UndirectedGraph(), right=tahini.core.UndirectedGraph())),
    # obj
    ([], dict(left=tahini.core.Graph(), right=tahini.core.Graph(), obj='ChildGraph')),
    ([], dict(left=tahini.core.UndirectedGraph(), right=tahini.core.UndirectedGraph(), obj='ChildGraph')),
    # non empty nodes
    ([], dict(left=tahini.core.Graph(order=1), right=tahini.core.Graph(order=1))),
    ([], dict(left=tahini.core.Graph(order=2), right=tahini.core.Graph(order=2))),
    ([], dict(left=tahini.core.UndirectedGraph(order=1), right=tahini.core.UndirectedGraph(order=1))),
    ([], dict(left=tahini.core.UndirectedGraph(order=2), right=tahini.core.UndirectedGraph(order=2))),
    # non empty edges
    ([], dict(left=tahini.core.Graph(edges=[(0, 1)]), right=tahini.core.Graph(edges=[(0, 1)]))),
    ([], dict(left=tahini.core.UndirectedGraph(edges=[(0, 1)]), right=tahini.core.UndirectedGraph(edges=[(0, 1)]))),
    # non empty edges with nodes extra
    (
        [],
        dict(
            left=tahini.core.Graph(edges=[(0, 1), (1, 2)]),
            right=tahini.core.Graph(nodes=[0, 1, 2], edges=[(0, 1), (1, 2)]),
        ),
    ),
    (
        [],
        dict(
            left=tahini.core.UndirectedGraph(edges=[(0, 1), (1, 2)]),
            right=tahini.core.UndirectedGraph(nodes=[0, 1, 2], edges=[(0, 1), (1, 2)]),
        ),
    ),
    # order does not matter for undirected graph
    (
        [],
        dict(
            left=tahini.core.UndirectedGraph(edges=[(0, 1)]),
            right=tahini.core.UndirectedGraph(edges=[(1, 0)]),
        ),
    ),
])
def test_assert_graph_equal(args, kwargs):
    tahini.testing.testing.assert_graph_equal(*args, **kwargs)



from __future__ import annotations
from collections.abc import Collection, Iterable
from typing import (
    Optional, Union, Dict, Any, Callable, TypeVar, Iterable as TypeIterable, Sequence as TypeSequence, Hashable,
    NoReturn,
)

from pandas import DataFrame, Series, Index, MultiIndex

__all__ = [
    'ContainerDataIndexed',
    'ContainerDataIndexedMulti',
    'ContainerDataIndexedMultiSets',
    'TypeContainerDataIndexed',
    'TypeIndexInput',
    'TypeIndexMultiInput',
    'TypeDataInput',
    'TypeMapper',
]

TypeContainerDataIndexed = TypeVar('TypeContainerDataIndexed', bound='ContainerDataIndexed')
TypeIndexInput = Union[TypeContainerDataIndexed, Index, TypeIterable[Hashable]]
TypeIndexMultiInput = Union[TypeContainerDataIndexed, MultiIndex, TypeIterable[TypeSequence[Hashable]]]
TypeDataInput = Union[DataFrame, Dict, Iterable]
TypeMapper = Union[Callable, Dict, Series]


class ContainerDataIndexed(Collection):

    _names_index = ['index']
    _name_index_internal = 'index_internal'

    def __init__(
            self,
            index: Optional[TypeIndexInput] = None,
            data: Optional[TypeDataInput] = None,
            **kwargs,
    ):

        if isinstance(index, self.__class__):
            data = index.data
            index = None

        if index is not None and not isinstance(index, Index):
            index = self._validate_index(index=index)

        if data is not None:
            data = DataFrame(data=data, **kwargs)
            if index is not None:
                data.index = index
        else:
            data = DataFrame(index=index, **kwargs)

        self.data = data

    @property
    def names_index(self) -> TypeSequence[str]:
        return self._names_index

    @property
    def data_internal(
            self,
    ) -> DataFrame:
        return self._data

    @property
    def data(
            self,
    ) -> DataFrame:
        return self.data_internal.set_index(keys=self._names_index)

    @data.setter
    def data(
            self,
            value: DataFrame,
    ) -> NoReturn:

        if len(value.columns.intersection(self._names_index)) == len(self._names_index):
            value = value.set_index(self._names_index)

        data = (
            self._validate_data(data=value)
            .assign(**{self._name_index_internal: lambda x: self._create_index_internal(x.index)})
            .reset_index()
            .set_index(self._name_index_internal)
        )

        self._data = data

    @property
    def data_testing(
            self,
    ) -> DataFrame:
        return self.data_internal.drop(columns=self._names_index)

    @classmethod
    def _validate_data(
            cls,
            data: DataFrame,
    ) -> DataFrame:
        data.index = cls._validate_index(index=data.index)
        return data

    @classmethod
    def _validate_index(
            cls,
            index: Union[TypeIndexInput, TypeIndexMultiInput],
    ) -> Union[Index, MultiIndex]:
        if not isinstance(index, Index):
            index = Index(index)
        index_internal = cls._create_index_internal(index=index)
        if not index_internal.is_unique:
            raise ValueError(f"Index needs to be unique for '{cls.__name__}'")
        index.names = cls._names_index
        return index

    @staticmethod
    def _create_index_internal(
            index: Union[Index, MultiIndex],
    ) -> Index:
        return index

    def drop(
            self,
            index: Optional[Union[TypeIndexInput, TypeIndexMultiInput]] = None,
            **kwargs,
    ) -> TypeContainerDataIndexed:
        if index is None:
            index = []
        self.data = self.data_internal.drop(index=self.__class__(index=index).data_internal.index, **kwargs)
        return self

    def update(
            self,
            index: Optional[Union[TypeIndexInput, TypeIndexMultiInput]] = None,
            data: Optional[TypeDataInput] = None,
            func: Optional[Callable] = None,
            **kwargs,
    ) -> TypeContainerDataIndexed:

        if func is None:
            func = self._update_func

        container = self.__class__(index=index, data=data)

        self.data = (
            self.data_internal
            .reindex(columns=self.data_internal.columns.union(container.data_internal.columns))
            .combine(
                other=container.data_internal,
                func=func,
                **kwargs,
            )
        )

        return self

    @staticmethod
    def _update_func(
            s1: Series,
            s2: Series,
    ) -> Series:
        return s2.fillna(s1)

    def map(
            self,
            mapper: Optional[TypeMapper] = None,
            **kwargs,
    ) -> TypeContainerDataIndexed:
        if mapper is not None:
            self.data = (
                self.data_internal
                .assign(**{
                    column: self.data_internal[column].map(mapper, **kwargs)
                    for column in self._names_index
                })
            )
        return self

    def __repr__(self):
        return f'{self.__class__.__name__}(index={self.data.index})'

    def iter(self) -> Iterable:
        return iter(self.data_internal.index)

    def __iter__(self):
        return self.iter()

    def __contains__(
            self,
            item: Any,
    ):
        return self.__class__(index=[item]).data_internal.index[0] in self.data_internal.index

    def __len__(self):
        return len(self.data_internal.index)

    def __eq__(
            self,
            other: Any,
    ):
        return (
            isinstance(other, self.__class__) and
            len(self.data_internal.index.symmetric_difference(other.data_internal.index)) == 0
        )


class ContainerDataIndexedMulti(ContainerDataIndexed):

    _names_index = ['index_0', 'index_1']

    def __init__(
            self,
            index: Optional[TypeIndexMultiInput] = None,
            **kwargs,
    ):
        super().__init__(index=index, **kwargs)

    @staticmethod
    def _create_index_internal(
            index: MultiIndex,
    ) -> Index:
        return index.to_flat_index()

    @classmethod
    def _validate_index(
            cls,
            index: Union[TypeIndexInput, TypeIndexMultiInput]
    ) -> MultiIndex:

        num_levels = len(cls._names_index)

        if len(index) == 0:
            index = MultiIndex(levels=[[]] * num_levels, codes=[[]] * num_levels)
        else:
            index = MultiIndex.from_tuples(index)

        index = super()._validate_index(index=index)

        return index


class ContainerDataIndexedMultiSets(ContainerDataIndexedMulti):

    @staticmethod
    def _create_index_internal(
            index: MultiIndex,
    ) -> Index:
        return index.to_flat_index().map(frozenset)



from functools import partial

import pytest
import pandas as pd

import tahini.core.nodes
import tahini.testing

assert_frame_equal = partial(
    pd.testing.assert_frame_equal,
    check_dtype=False,
    check_column_type=False,
    check_index_type=False,
)


@pytest.mark.parametrize('args, kwargs', [
    # empty
    ([], dict()),
    # order
    ([], dict(order=1)),
])
def test_nodes_init_simple(args, kwargs):
    tahini.core.nodes.Nodes(*args, **kwargs)


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty
    ([], dict(), tahini.core.nodes.Nodes()),
    # order
    ([], dict(order=1), tahini.core.nodes.Nodes(index=[0])),
    ([], dict(order=2), tahini.core.nodes.Nodes(index=[0, 1])),
])
def test_nodes_init(args, kwargs, expected):
    nodes = tahini.core.nodes.Nodes(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(nodes, expected)


@pytest.mark.parametrize('nodes, args, kwargs, expected', [
    # empty
    (
        tahini.core.nodes.Nodes(),
        [],
        dict(),
        pd.DataFrame(columns=['position_dim_0', 'position_dim_1'], index=pd.Index([], name='node')),
    ),
])
def test_nodes_get_positions(nodes, args, kwargs, expected):
    positions = nodes.get_positions(*args, **kwargs)
    assert_frame_equal(positions, expected)



from functools import partial

import pytest
import pandas as pd

import tahini.core.graph
import tahini.core.edges
import tahini.core.nodes
import tahini.testing


names_edges = tahini.core.graph.Graph().edges.names_index
name_nodes = tahini.core.graph.Graph().nodes.names_index[0]

assert_frame_equal = partial(
    pd.testing.assert_frame_equal,
    check_column_type=False,
    check_index_type=False,
    check_dtype=False,
)


@pytest.mark.parametrize('args, kwargs, nodes_expected, edges_expected', [
    # empty
    ([], dict(), tahini.core.nodes.Nodes(), tahini.core.edges.Edges()),
    # nodes
    ([], dict(nodes=[0, 1]), tahini.core.nodes.Nodes(index=[0, 1]), tahini.core.edges.Edges()),
    # edges
    (
        [],
        dict(edges=[(0, 1), (1, 2)]),
        tahini.core.nodes.Nodes(index=[0, 1, 2]),
        tahini.core.edges.Edges(index=[(0, 1), (1, 2)]),
    ),
    # nodes and edges
    (
        [],
        dict(nodes=[0, 1, 2], edges=[(0, 1), (1, 2)]),
        tahini.core.nodes.Nodes(index=[0, 1, 2]),
        tahini.core.edges.Edges(index=[(0, 1), (1, 2)]),
    ),
    # partial nodes and edges
    (
        [],
        dict(nodes=[0, 1], edges=[(0, 1), (1, 2)]),
        tahini.core.nodes.Nodes(index=[0, 1, 2]),
        tahini.core.edges.Edges(index=[(0, 1), (1, 2)]),
    ),
    # order
    ([], dict(order=2), tahini.core.nodes.Nodes(index=range(2)), tahini.core.edges.Edges()),
    # nodes data
    (
        [],
        dict(nodes_data=pd.DataFrame(data=dict(value=['a', 'b']))),
        tahini.core.nodes.Nodes(data=dict(value=['a', 'b'])),
        tahini.core.edges.Edges(),
    ),
    # edges data
    (
        [],
        dict(edges_data=pd.DataFrame(data=dict(value=['a', 'b']), index=[(0, 1), (1, 2)])),
        tahini.core.nodes.Nodes(index=[0, 1, 2]),
        tahini.core.edges.Edges(data=dict(value=['a', 'b']), index=[(0, 1), (1, 2)]),
    ),
])
def test_graph_init(args, kwargs, nodes_expected, edges_expected):
    graph = tahini.core.graph.Graph(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(graph.nodes, nodes_expected)
    tahini.testing.testing.assert_container_equal(graph.edges, edges_expected)


@pytest.mark.parametrize('args, kwargs, nodes_expected, edges_expected', [
    # empty
    ([], dict(), tahini.core.nodes.Nodes(), tahini.core.edges.UndirectedEdges()),
    # nodes
    ([], dict(nodes=[0, 1]), tahini.core.nodes.Nodes(index=[0, 1]), tahini.core.edges.UndirectedEdges()),
    # edges
    (
        [],
        dict(edges=[(0, 1), (1, 2)]),
        tahini.core.nodes.Nodes(index=[0, 1, 2]),
        tahini.core.edges.UndirectedEdges(index=[(0, 1), (1, 2)]),
    ),
    # nodes and edges
    (
        [],
        dict(nodes=[0, 1, 2], edges=[(0, 1), (1, 2)]),
        tahini.core.nodes.Nodes(index=[0, 1, 2]),
        tahini.core.edges.UndirectedEdges(index=[(0, 1), (1, 2)]),
    ),
    # partial nodes and edges
    (
        [],
        dict(nodes=[0, 1], edges=[(0, 1), (1, 2)]),
        tahini.core.nodes.Nodes(index=[0, 1, 2]),
        tahini.core.edges.UndirectedEdges(index=[(0, 1), (1, 2)]),
    ),
    # order
    ([], dict(order=2), tahini.core.nodes.Nodes(index=range(2)), tahini.core.edges.UndirectedEdges()),
    # nodes data
    (
        [],
        dict(nodes_data=pd.DataFrame(data=dict(value=['a', 'b']))),
        tahini.core.nodes.Nodes(data=dict(value=['a', 'b'])),
        tahini.core.edges.UndirectedEdges(),
    ),
    # edges data
    (
        [],
        dict(edges_data=pd.DataFrame(data=dict(value=['a', 'b']), index=[(0, 1), (1, 2)])),
        tahini.core.nodes.Nodes(index=[0, 1, 2]),
        tahini.core.edges.UndirectedEdges(data=dict(value=['a', 'b']), index=[(0, 1), (1, 2)]),
    ),
])
def test_undirected_graph_init(args, kwargs, nodes_expected, edges_expected):
    graph = tahini.core.graph.UndirectedGraph(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(graph.nodes, nodes_expected)
    tahini.testing.testing.assert_container_equal(graph.edges, edges_expected)


@pytest.mark.parametrize('klass, attribute, expected', [
    # _type_edges
    (tahini.core.graph.Graph, '_type_edges', tahini.core.edges.Edges),
    (tahini.core.graph.UndirectedGraph, '_type_edges', tahini.core.edges.UndirectedEdges),
    # _columns_degree_keep
    (tahini.core.graph.Graph, '_columns_degree_keep', ['degree_in', 'degree_out', 'degree']),
    (tahini.core.graph.UndirectedGraph, '_columns_degree_keep', ['degree']),
    # _columns_neighbors_keep
    (tahini.core.graph.Graph, '_columns_neighbors_keep', ['neighbors_in', 'neighbors_out', 'neighbors']),
    (tahini.core.graph.UndirectedGraph, '_columns_neighbors_keep', ['neighbors']),
])
def test_graph_attribute(klass, attribute, expected):
    value = getattr(klass, attribute)
    assert value == expected


@pytest.mark.parametrize('klass, args, kwargs, expected', [
    # trivial order
    (tahini.core.graph.Graph, [], dict(order=0), tahini.core.graph.Graph()),
    (tahini.core.graph.Graph, [], dict(order=1), tahini.core.graph.Graph(order=1)),
    (tahini.core.graph.UndirectedGraph, [], dict(order=0), tahini.core.graph.UndirectedGraph()),
    (tahini.core.graph.UndirectedGraph, [], dict(order=1), tahini.core.graph.UndirectedGraph(order=1)),
    # non trivial order
    (tahini.core.graph.Graph, [], dict(order=2), tahini.core.graph.Graph(edges=[(0, 1)])),
    (tahini.core.graph.Graph, [], dict(order=3), tahini.core.graph.Graph(edges=[(0, 1), (1, 2)])),
    (tahini.core.graph.UndirectedGraph, [], dict(order=2), tahini.core.graph.UndirectedGraph(edges=[(0, 1)])),
    (tahini.core.graph.UndirectedGraph, [], dict(order=3), tahini.core.graph.UndirectedGraph(edges=[(0, 1), (1, 2)])),
    # order and nodes
    (tahini.core.graph.Graph, [], dict(order=2, nodes=['a', 'b']), tahini.core.graph.Graph(edges=[('a', 'b')])),
    (
        tahini.core.graph.Graph,
        [],
        dict(order=3, nodes=['a', 'b', 'c']),
        tahini.core.graph.Graph(edges=[('a', 'b'), ('b', 'c')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=2, nodes=['a', 'b']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=3, nodes=['a', 'b', 'c']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b'), ('b', 'c')]),
    ),
    # trivial nodes
    (tahini.core.graph.Graph, [], dict(nodes=[]), tahini.core.graph.Graph()),
    (tahini.core.graph.Graph, [], dict(nodes=['a']), tahini.core.graph.Graph(nodes=['a'])),
    (tahini.core.graph.UndirectedGraph, [], dict(nodes=[]), tahini.core.graph.UndirectedGraph()),
    (tahini.core.graph.UndirectedGraph, [], dict(nodes=['a']), tahini.core.graph.UndirectedGraph(nodes=['a'])),
    # non trivial nodes
    (tahini.core.graph.Graph, [], dict(nodes=['a', 'b']), tahini.core.graph.Graph(edges=[('a', 'b')])),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(nodes=['a', 'b']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b')]),
    ),
])
def test_graph_path(klass, args, kwargs, expected):
    graph = klass.path(*args, **kwargs)
    tahini.testing.testing.assert_graph_equal(graph, expected)


@pytest.mark.parametrize('klass, args, kwargs, type_error, message_error', [
    # trivial order
    (
        tahini.core.graph.Graph,
        [],
        dict(order=0),
        ValueError,
        "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    (
        tahini.core.graph.Graph,
        [],
        dict(order=1),
        ValueError,
        "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    (
        tahini.core.graph.Graph,
        [],
        dict(order=2),
        ValueError,
        "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=0),
        ValueError,
        "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=1),
        ValueError,
        "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=2),
        ValueError,
        "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    # trivial nodes
    (
        tahini.core.graph.Graph,
        [], dict(nodes=[]),
        ValueError,
        "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    (
        tahini.core.graph.Graph,
        [],
        dict(nodes=[0]),
        ValueError,
        "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    (
        tahini.core.graph.Graph,
        [],
        dict(nodes=[0, 1]),
        ValueError,
        "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    (
            tahini.core.graph.UndirectedGraph,
            [],
            dict(nodes=[]),
            ValueError,
            "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    (
            tahini.core.graph.UndirectedGraph,
            [],
            dict(nodes=[0]),
            ValueError,
            "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
    (
            tahini.core.graph.UndirectedGraph,
            [],
            dict(nodes=[0, 1]),
            ValueError,
            "Inputs 'order' or length of 'nodes' has to be >= 3 for cycle",
    ),
])
def test_graph_cycle_error(klass, args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        klass.cycle(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('klass, args, kwargs, expected', [
    # non trivial order
    (tahini.core.graph.Graph, [], dict(order=3), tahini.core.graph.Graph(edges=[(0, 1), (1, 2), (2, 0)])),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=3),
        tahini.core.graph.UndirectedGraph(edges=[(0, 1), (1, 2), (2, 0)]),
    ),
    # order and nodes
    (
        tahini.core.graph.Graph,
        [],
        dict(order=3, nodes=['a', 'b', 'c']),
        tahini.core.graph.Graph(edges=[('a', 'b'), ('b', 'c'), ('c', 'a')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=3, nodes=['a', 'b', 'c']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b'), ('b', 'c'), ('c', 'a')]),
    ),
    # non trivial nodes
    (
        tahini.core.graph.Graph,
        [],
        dict(nodes=['a', 'b', 'c']),
        tahini.core.graph.Graph(edges=[('a', 'b'), ('b', 'c'), ('c', 'a')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(nodes=['a', 'b', 'c']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b'), ('b', 'c'), ('c', 'a')]),
    ),
])
def test_graph_cycle(klass, args, kwargs, expected):
    graph = klass.cycle(*args, **kwargs)
    tahini.testing.testing.assert_graph_equal(graph, expected)


@pytest.mark.parametrize('klass, args, kwargs, expected', [
    # trivial order
    (tahini.core.graph.Graph, [], dict(order=0), tahini.core.graph.Graph()),
    (tahini.core.graph.Graph, [], dict(order=1), tahini.core.graph.Graph(order=1)),
    (tahini.core.graph.UndirectedGraph, [], dict(order=0), tahini.core.graph.UndirectedGraph()),
    (tahini.core.graph.UndirectedGraph, [], dict(order=1), tahini.core.graph.UndirectedGraph(order=1)),
    # non trivial order
    (tahini.core.graph.Graph, [], dict(order=2), tahini.core.graph.Graph(edges=[(0, 1)])),
    (tahini.core.graph.Graph, [], dict(order=3), tahini.core.graph.Graph(edges=[(0, 1), (0, 2)])),
    (tahini.core.graph.UndirectedGraph, [], dict(order=2), tahini.core.graph.UndirectedGraph(edges=[(0, 1)])),
    (tahini.core.graph.UndirectedGraph, [], dict(order=3), tahini.core.graph.UndirectedGraph(edges=[(0, 1), (0, 2)])),
    # order and nodes
    (tahini.core.graph.Graph, [], dict(order=2, nodes=['a', 'b']), tahini.core.graph.Graph(edges=[('a', 'b')])),
    (
        tahini.core.graph.Graph,
        [],
        dict(order=3, nodes=['a', 'b', 'c']),
        tahini.core.graph.Graph(edges=[('a', 'b'), ('a', 'c')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=2, nodes=['a', 'b']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=3, nodes=['a', 'b', 'c']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b'), ('a', 'c')]),
    ),
    # trivial nodes
    (tahini.core.graph.Graph, [], dict(nodes=[]), tahini.core.graph.Graph()),
    (tahini.core.graph.Graph, [], dict(nodes=['a']), tahini.core.graph.Graph(nodes=['a'])),
    (tahini.core.graph.UndirectedGraph, [], dict(nodes=[]), tahini.core.graph.UndirectedGraph()),
    (tahini.core.graph.UndirectedGraph, [], dict(nodes=['a']), tahini.core.graph.UndirectedGraph(nodes=['a'])),
    # non trivial nodes
    (tahini.core.graph.Graph, [], dict(nodes=['a', 'b']), tahini.core.graph.Graph(edges=[('a', 'b')])),
    (tahini.core.graph.Graph, [], dict(nodes=['a', 'b', 'c']), tahini.core.graph.Graph(edges=[('a', 'b'), ('a', 'c')])),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(nodes=['a', 'b']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(nodes=['a', 'b', 'c']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b'), ('a', 'c')]),
    ),
])
def test_graph_star(klass, args, kwargs, expected):
    graph = klass.star(*args, **kwargs)
    tahini.testing.testing.assert_graph_equal(graph, expected)


@pytest.mark.parametrize('klass, args, kwargs, expected', [
    # trivial order
    (tahini.core.graph.Graph, [], dict(order=0), tahini.core.graph.Graph()),
    (tahini.core.graph.Graph, [], dict(order=1), tahini.core.graph.Graph(order=1)),
    (tahini.core.graph.UndirectedGraph, [], dict(order=0), tahini.core.graph.UndirectedGraph()),
    (tahini.core.graph.UndirectedGraph, [], dict(order=1), tahini.core.graph.UndirectedGraph(order=1)),
    # non trivial order
    (tahini.core.graph.Graph, [], dict(order=2), tahini.core.graph.Graph(edges=[(0, 1), (1, 0)])),
    (
        tahini.core.graph.Graph,
        [], dict(order=3),
        tahini.core.graph.Graph(edges=[(0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1)]),
    ),
    (tahini.core.graph.UndirectedGraph, [], dict(order=2), tahini.core.graph.UndirectedGraph(edges=[(0, 1)])),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=3),
        tahini.core.graph.UndirectedGraph(edges=[(0, 1), (0, 2), (1, 2)]),
    ),
    # order and nodes
    (
        tahini.core.graph.Graph,
        [],
        dict(order=2, nodes=['a', 'b']),
        tahini.core.graph.Graph(edges=[('a', 'b'), ('b', 'a')]),
    ),
    (
        tahini.core.graph.Graph,
        [],
        dict(order=3, nodes=['a', 'b', 'c']),
        tahini.core.graph.Graph(edges=[('a', 'b'), ('a', 'c'), ('b', 'a'), ('b', 'c'), ('c', 'a'), ('c', 'b')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=2, nodes=['a', 'b']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(order=3, nodes=['a', 'b', 'c']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b'), ('a', 'c'), ('b', 'c')]),
    ),
    # trivial nodes
    (tahini.core.graph.Graph, [], dict(nodes=[]), tahini.core.graph.Graph()),
    (tahini.core.graph.Graph, [], dict(nodes=['a']), tahini.core.graph.Graph(nodes=['a'])),
    (tahini.core.graph.UndirectedGraph, [], dict(nodes=[]), tahini.core.graph.UndirectedGraph()),
    (tahini.core.graph.UndirectedGraph, [], dict(nodes=['a']), tahini.core.graph.UndirectedGraph(nodes=['a'])),
    # non trivial nodes
    (tahini.core.graph.Graph, [], dict(nodes=['a', 'b']), tahini.core.graph.Graph(edges=[('a', 'b'), ('b', 'a')])),
    (
        tahini.core.graph.Graph,
        [],
        dict(nodes=['a', 'b', 'c']),
        tahini.core.graph.Graph(edges=[('a', 'b'), ('a', 'c'), ('b', 'a'), ('b', 'c'), ('c', 'a'), ('c', 'b')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(nodes=['a', 'b']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b')]),
    ),
    (
        tahini.core.graph.UndirectedGraph,
        [],
        dict(nodes=['a', 'b', 'c']),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b'), ('a', 'c'), ('b', 'c')]),
    ),
])
def test_graph_complete(klass, args, kwargs, expected):
    graph = klass.complete(*args, **kwargs)
    tahini.testing.testing.assert_graph_equal(graph, expected)


@pytest.mark.parametrize('graph, args, kwargs, expected', [
    # empty
    (tahini.core.graph.Graph(), [], dict(), tahini.core.graph.Graph()),
    (tahini.core.graph.Graph(), [], dict(nodes=lambda x: x.nodes), tahini.core.graph.Graph()),
    (tahini.core.graph.Graph(), [], dict(nodes=lambda x: x.edges), tahini.core.graph.Graph()),
    # non empty inputs
    (tahini.core.graph.Graph(), [], dict(nodes=[0]), tahini.core.graph.Graph(nodes=[0])),
    (tahini.core.graph.Graph(), [], dict(edges=[(0, 1)]), tahini.core.graph.Graph(edges=[(0, 1)])),
    # non empty graph
    (tahini.core.graph.Graph(nodes=[0]), [], dict(nodes=[1]), tahini.core.graph.Graph(nodes=[1])),
    (
        tahini.core.graph.Graph(edges=[(0, 1)]),
        [],
        dict(edges=[(0, 2)]),
        tahini.core.graph.Graph(nodes=[1], edges=[(0, 2)]),
    ),
])
def test_graph_assign(graph, args, kwargs, expected):
    graph_assigned = graph.assign(*args, **kwargs)
    tahini.testing.testing.assert_graph_equal(graph_assigned, expected)


@pytest.mark.parametrize('graph, edges, args, kwargs, nodes_expected, graph_expected', [
    # empty
    (
        tahini.core.graph.Graph(),
        tahini.core.edges.Edges(index=[(0, 1)]),
        [],
        dict(),
        tahini.core.nodes.Nodes(index=[0, 1]),
        tahini.core.graph.Graph(edges=[(0, 1)]),
    ),
    (
        tahini.core.graph.UndirectedGraph(),
        tahini.core.edges.UndirectedEdges(index=[(0, 1)]),
        [],
        dict(),
        tahini.core.nodes.Nodes(index=[0, 1]),
        tahini.core.graph.UndirectedGraph(edges=[(0, 1)]),
    ),
    # non empty
    (
        tahini.core.graph.Graph(nodes=[0]),
        tahini.core.edges.Edges(index=[(1, 2)]),
        [],
        dict(),
        tahini.core.nodes.Nodes(index=[0, 1, 2]),
        tahini.core.graph.Graph(nodes=[0, 1, 2], edges=[(1, 2)]),
    ),
    (
        tahini.core.graph.UndirectedGraph(nodes=[0]),
        tahini.core.edges.UndirectedEdges(index=[(1, 2)]),
        [],
        dict(),
        tahini.core.nodes.Nodes(index=[0, 1, 2]),
        tahini.core.graph.UndirectedGraph(nodes=[0, 1, 2], edges=[(1, 2)]),
    ),
])
def test_graph__update_nodes_from_edges(graph, edges, args, kwargs, nodes_expected, graph_expected):
    graph._edges = edges
    nodes = graph._update_nodes_from_edges(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(nodes, nodes_expected)
    tahini.testing.testing.assert_graph_equal(graph, graph_expected)


@pytest.mark.parametrize('graph, args, kwargs, type_error, message_error', [
    # non unique update
    (tahini.core.graph.Graph(), [], dict(nodes=[0, 0]), ValueError, "Index needs to be unique for 'Nodes'"),
])
def test_graph_update_nodes_error(graph, args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        graph.update_nodes(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('graph, args, kwargs, expected', [
    # empty
    (tahini.core.graph.Graph(), [], dict(), tahini.core.graph.Graph()),
    (tahini.core.graph.UndirectedGraph(), [], dict(), tahini.core.graph.UndirectedGraph()),
    # non empty graph
    (tahini.core.graph.Graph(nodes=[0, 1]), [], dict(), tahini.core.graph.Graph(nodes=[0, 1])),
    (tahini.core.graph.UndirectedGraph(nodes=[0, 1]), [], dict(), tahini.core.graph.UndirectedGraph(nodes=[0, 1])),
    # non empty nodes
    (tahini.core.graph.Graph(), [], dict(nodes=[0, 1]), tahini.core.graph.Graph(nodes=[0, 1])),
    (tahini.core.graph.UndirectedGraph(), [], dict(nodes=[0, 1]), tahini.core.graph.UndirectedGraph(nodes=[0, 1])),
    # non empty both same
    (tahini.core.graph.Graph(nodes=[0, 1]), [], dict(nodes=[0, 1]), tahini.core.graph.Graph(nodes=[0, 1])),
    (
        tahini.core.graph.UndirectedGraph(nodes=[0, 1]),
        [],
        dict(nodes=[0, 1]),
        tahini.core.graph.UndirectedGraph(nodes=[0, 1]),
    ),
    # non empty both different
    (tahini.core.graph.Graph(nodes=[0, 1]), [], dict(nodes=[2]), tahini.core.graph.Graph(nodes=[0, 1, 2])),
    (
        tahini.core.graph.UndirectedGraph(nodes=[0, 1]),
        [],
        dict(nodes=[2]),
        tahini.core.graph.UndirectedGraph(nodes=[0, 1, 2]),
    ),
    # data
    (
        tahini.core.graph.Graph(nodes=[0, 1]),
        [],
        dict(data=dict(value=['a', 'b'])),
        tahini.core.graph.Graph(nodes_data=dict(value=['a', 'b']), nodes=[0, 1]),
    ),
    (
        tahini.core.graph.UndirectedGraph(nodes=[0, 1]),
        [],
        dict(data=dict(value=['a', 'b'])),
        tahini.core.graph.UndirectedGraph(nodes_data=dict(value=['a', 'b']), nodes=[0, 1]),
    ),
])
def test_graph_update_nodes(graph, args, kwargs, expected):
    graph_updated = graph.update_nodes(*args, **kwargs)
    tahini.testing.testing.assert_graph_equal(graph_updated, expected)


@pytest.mark.parametrize('graph, nodes, args, kwargs, edges_expected, graph_expected', [
    # empty
    (
        tahini.core.graph.Graph(),
        tahini.core.nodes.Nodes(index=[0, 1]),
        [],
        dict(),
        tahini.core.edges.Edges(),
        tahini.core.graph.Graph(nodes=[0, 1]),
    ),
    (
        tahini.core.graph.UndirectedGraph(),
        tahini.core.nodes.Nodes(index=[0, 1]),
        [],
        dict(),
        tahini.core.edges.UndirectedEdges(),
        tahini.core.graph.UndirectedGraph(nodes=[0, 1]),
    ),
    # non empty
    (
        tahini.core.graph.Graph(edges=[(0, 1), (1, 2)]),
        tahini.core.nodes.Nodes(index=[0, 1]),
        [],
        dict(),
        tahini.core.edges.Edges(index=[(0, 1)]),
        tahini.core.graph.Graph(edges=[(0, 1)]),
    ),
    (
        tahini.core.graph.UndirectedGraph(edges=[(0, 1), (1, 2)]),
        tahini.core.nodes.Nodes(index=[0, 1]),
        [],
        dict(),
        tahini.core.edges.UndirectedEdges(index=[(0, 1)]),
        tahini.core.graph.UndirectedGraph(edges=[(0, 1)]),
    ),
])
def test_graph__update_edges_from_nodes(graph, nodes, args, kwargs, edges_expected, graph_expected):
    graph._nodes = nodes
    edges = graph._update_edges_from_nodes(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(edges, edges_expected)
    tahini.testing.testing.assert_graph_equal(graph, graph_expected)


@pytest.mark.parametrize('graph, args, kwargs, type_error, message_error', [
    # non unique update
    (tahini.core.graph.Graph(), [], dict(edges=[(0, 1), (0, 1)]), ValueError, "Index needs to be unique for 'Edges'"),
    (
        tahini.core.graph.UndirectedGraph(),
        [],
        dict(edges=[(0, 1), (1, 0)]),
        ValueError,
        "Index needs to be unique for 'UndirectedEdges'",
    ),
])
def test_graph_update_edges_error(graph, args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        graph.update_edges(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('graph, args, kwargs, expected', [
    # empty
    (tahini.core.graph.Graph(), [], dict(), tahini.core.graph.Graph()),
    (tahini.core.graph.UndirectedGraph(), [], dict(), tahini.core.graph.UndirectedGraph()),
    # non empty graph
    (tahini.core.graph.Graph(edges=[(0, 1)]), [], dict(), tahini.core.graph.Graph(edges=[(0, 1)])),
    (tahini.core.graph.UndirectedGraph(edges=[(0, 1)]), [], dict(), tahini.core.graph.UndirectedGraph(edges=[(0, 1)])),
    # non empty edges
    (tahini.core.graph.Graph(), [], dict(edges=[(0, 1)]), tahini.core.graph.Graph(edges=[(0, 1)])),
    (tahini.core.graph.UndirectedGraph(), [], dict(edges=[(0, 1)]), tahini.core.graph.UndirectedGraph(edges=[(0, 1)])),
    # non empty both different
    (tahini.core.graph.Graph(edges=[(0, 1)]), [], dict(edges=[(0, 1)]), tahini.core.graph.Graph(edges=[(0, 1)])),
    (
        tahini.core.graph.UndirectedGraph(edges=[(0, 1)]),
        [],
        dict(edges=[(0, 1)]),
        tahini.core.graph.UndirectedGraph(edges=[(0, 1)]),
    ),
    # non empty both different
    (
        tahini.core.graph.Graph(edges=[(0, 1)]),
        [],
        dict(edges=[(1, 2)]),
        tahini.core.graph.Graph(edges=[(0, 1), (1, 2)]),
    ),
    (
        tahini.core.graph.UndirectedGraph(edges=[(0, 1)]),
        [],
        dict(edges=[(1, 2)]),
        tahini.core.graph.UndirectedGraph(edges=[(1, 2), (0, 1)]),
    ),
    # data
    (
        tahini.core.graph.Graph(edges=[(0, 1), (1, 2)]),
        [],
        dict(edges=[(0, 1), (1, 2)], data=dict(value=['a', 'b'])),
        tahini.core.graph.Graph(edges_data=dict(value=['a', 'b']), edges=[(0, 1), (1, 2)]),
    ),
    (
        tahini.core.graph.UndirectedGraph(edges=[(0, 1), (1, 2)]),
        [],
        dict(edges=[(0, 1), (1, 2)], data=dict(value=['a', 'b'])),
        tahini.core.graph.UndirectedGraph(edges_data=dict(value=['a', 'b']), edges=[(0, 1), (1, 2)]),
    ),
])
def test_graph_update_edges(graph, args, kwargs, expected):
    graph_updated = graph.update_edges(*args, **kwargs)
    tahini.testing.testing.assert_graph_equal(graph_updated, expected)


@pytest.mark.parametrize('graph, args, kwargs, expected', [
    # empty all
    (tahini.core.graph.Graph(), [], dict(), tahini.core.graph.Graph()),
    (tahini.core.graph.UndirectedGraph(), [], dict(), tahini.core.graph.UndirectedGraph()),
    # empty list
    (tahini.core.graph.Graph(), [], dict(nodes=[]), tahini.core.graph.Graph()),
    (tahini.core.graph.UndirectedGraph(), [], dict(nodes=[]), tahini.core.graph.UndirectedGraph()),
    # empty inputs
    (tahini.core.graph.Graph(nodes=[0, 1]), [], dict(), tahini.core.graph.Graph(nodes=[0, 1])),
    (tahini.core.graph.UndirectedGraph(nodes=[0, 1]), [], dict(), tahini.core.graph.UndirectedGraph(nodes=[0, 1])),
    # non empty
    (tahini.core.graph.Graph(nodes=[0, 1]), [], dict(nodes=[0]), tahini.core.graph.Graph(nodes=[1])),
    (
        tahini.core.graph.UndirectedGraph(nodes=[0, 1]),
        [],
        dict(nodes=[0]),
        tahini.core.graph.UndirectedGraph(nodes=[1]),
    ),
    # non empty with removing edges
    (
        tahini.core.graph.Graph(nodes=[0, 1], edges=[(0, 1)]),
        [],
        dict(nodes=[0]),
        tahini.core.graph.Graph(nodes=[1], edges=tahini.core.edges.Edges()),
    ),
    (
        tahini.core.graph.UndirectedGraph(nodes=[0, 1], edges=[(0, 1)]),
        [],
        dict(nodes=[0]),
        tahini.core.graph.UndirectedGraph(nodes=[1], edges=tahini.core.edges.Edges()),
    ),
])
def test_graph_drop_nodes(graph, args, kwargs, expected):
    graph_dropped = graph.drop_nodes(*args, **kwargs)
    tahini.testing.testing.assert_graph_equal(graph_dropped, expected)


@pytest.mark.parametrize('graph, args, kwargs, expected', [
    # empty all
    (tahini.core.graph.Graph(), [], dict(), tahini.core.graph.Graph()),
    (tahini.core.graph.UndirectedGraph(), [], dict(), tahini.core.graph.UndirectedGraph()),
    # empty list
    (tahini.core.graph.Graph(), [], dict(edges=[]), tahini.core.graph.Graph()),
    (tahini.core.graph.UndirectedGraph(), [], dict(edges=[]), tahini.core.graph.UndirectedGraph()),
    # empty inputs
    (
        tahini.core.graph.Graph(edges=[(0, 1), (1, 2)]),
        [],
        dict(),
        tahini.core.graph.Graph(edges=[(0, 1), (1, 2)]),
    ),
    (
        tahini.core.graph.UndirectedGraph(edges=[(0, 1), (1, 2)]),
        [],
        dict(),
        tahini.core.graph.UndirectedGraph(edges=[(0, 1), (1, 2)]),
    ),
    # non empty
    (
        tahini.core.graph.Graph(edges=[(0, 1), (1, 2)]),
        [],
        dict(edges=[(0, 1)]),
        tahini.core.graph.Graph(nodes=[0, 1, 2], edges=[(1, 2)]),
    ),
    (
        tahini.core.graph.UndirectedGraph(edges=[(0, 1), (1, 2)]),
        [],
        dict(edges=[(0, 1)]),
        tahini.core.graph.UndirectedGraph(nodes=[0, 1, 2], edges=[(1, 2)]),
    ),
])
def test_graph_drop_edges(graph, args, kwargs, expected):
    graph_dropped = graph.drop_edges(*args, **kwargs)
    tahini.testing.testing.assert_graph_equal(graph_dropped, expected)


@pytest.mark.parametrize('graph, expected', [
    # empty
    (tahini.core.graph.Graph(), 0),
    (tahini.core.graph.UndirectedGraph(), 0),
    # non empty
    (tahini.core.graph.Graph(nodes=[0, 1]), 2),
    (tahini.core.graph.UndirectedGraph(nodes=[0, 1]), 2),
    # order
    (tahini.core.graph.Graph(order=3), 3),
    (tahini.core.graph.UndirectedGraph(order=3), 3),
])
def test_graph_order(graph, expected):
    order = graph.order
    assert order == expected


@pytest.mark.parametrize('graph, expected', [
    # empty
    (tahini.core.graph.Graph(), 0),
    (tahini.core.graph.UndirectedGraph(), 0),
    # non empty nodes
    (tahini.core.graph.Graph(nodes=[0, 1]), 0),
    (tahini.core.graph.UndirectedGraph(nodes=[0, 1]), 0),
    # order
    (tahini.core.graph.Graph(order=3), 0),
    (tahini.core.graph.UndirectedGraph(order=3), 0),
    # non empty edges
    (tahini.core.graph.Graph(edges=[(0, 1)]), 1),
    (tahini.core.graph.UndirectedGraph(edges=[(0, 1)]), 1),
    # non empty edges
    (tahini.core.graph.Graph(edges=[(0, 1), (1, 2)]), 2),
    (tahini.core.graph.UndirectedGraph(edges=[(0, 1), (1, 2)]), 2),
])
def test_graph_size(graph, expected):
    size = graph.size
    assert size == expected


@pytest.mark.parametrize('graph, expected', [
    # empty
    (tahini.core.graph.Graph(), f'Graph(nodes={tahini.core.nodes.Nodes()}, edges={tahini.core.edges.Edges()})'),
    (
        tahini.core.graph.UndirectedGraph(),
        f'UndirectedGraph(nodes={tahini.core.nodes.Nodes()}, edges={tahini.core.edges.UndirectedEdges()})',
    ),
])
def test_graph_repr(graph, expected):
    repr_graph = repr(graph)
    assert repr_graph == expected


@pytest.mark.parametrize('graph, args, kwargs, type_error, message_error', [
    # non unique mapping
    (
        tahini.core.graph.Graph(edges=[(0, 1), (0, 2)]),
        [],
        dict(mapper={1: 'a', 2: 'a'}),
        ValueError,
        "Index needs to be unique for 'Nodes'",
    ),
    (
        tahini.core.graph.UndirectedGraph(edges=[(0, 1), (0, 2)]),
        [],
        dict(mapper={1: 'a', 2: 'a'}),
        ValueError,
        "Index needs to be unique for 'Nodes'",
    ),
])
def test_graph_map_nodes_error(graph, args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        graph.map_nodes(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('graph, args, kwargs, expected', [
    # empty
    (tahini.core.graph.Graph(), [], dict(), tahini.core.graph.Graph()),
    (tahini.core.graph.UndirectedGraph(), [], dict(), tahini.core.graph.UndirectedGraph()),
    # empty mapping
    (tahini.core.graph.Graph(), [], dict(mapper=dict()), tahini.core.graph.Graph()),
    (tahini.core.graph.UndirectedGraph(), [], dict(mapper=dict()), tahini.core.graph.UndirectedGraph()),
    # non empty mapping
    (tahini.core.graph.Graph(), [], dict(mapper={}), tahini.core.graph.Graph()),
    (tahini.core.graph.UndirectedGraph(), [], dict(mapper={}), tahini.core.graph.UndirectedGraph()),
    # non empty graph
    (tahini.core.graph.Graph(order=1), [], dict(mapper={0: 'a'}), tahini.core.graph.Graph(nodes=['a'])),
    (
        tahini.core.graph.UndirectedGraph(order=1),
        [],
        dict(mapper={0: 'a'}),
        tahini.core.graph.UndirectedGraph(nodes=['a']),
    ),
    # non empty graph
    (
        tahini.core.graph.Graph(edges=[(0, 1)]),
        [],
        dict(mapper={0: 'a', 1: 'b'}),
        tahini.core.graph.Graph(edges=[('a', 'b')]),
    ),
    (
        tahini.core.graph.UndirectedGraph(edges=[(0, 1)]),
        [],
        dict(mapper={0: 'a', 1: 'b'}),
        tahini.core.graph.UndirectedGraph(edges=[('a', 'b')]),
    ),
])
def test_graph_map_nodes(graph, args, kwargs, expected):
    graph_mapped = graph.map_nodes(*args, **kwargs)
    tahini.testing.testing.assert_graph_equal(graph_mapped, expected)


@pytest.mark.parametrize('graph, args, kwargs, expected', [
    # empty
    (
        tahini.core.graph.Graph(),
        [],
        dict(column_name='x'),
        pd.DataFrame(columns=['x'], index=pd.Index([], name=name_nodes)),
    ),
    # no edges
    (
        tahini.core.graph.Graph(nodes=[0]),
        [],
        dict(column_name='x'),
        pd.DataFrame(data=dict(x=[0]), index=pd.Index([0], name=name_nodes)),
    ),
    (
        tahini.core.graph.Graph(nodes=[0, 1]),
        [],
        dict(column_name='x'),
        pd.DataFrame(data=dict(x=[0, 0]), index=pd.Index([0, 1], name=name_nodes)),
    ),
    # edges
    (
        tahini.core.graph.Graph(edges=[(0, 1)]),
        [],
        dict(column_name='x'),
        pd.DataFrame(data=dict(x=[1, 1]), index=pd.Index([0, 1], name=name_nodes)),
    ),
    (
        tahini.core.graph.Graph(edges=[(0, 1), (0, 2)]),
        [],
        dict(column_name='x'),
        pd.DataFrame(data=dict(x=[2, 1, 1]), index=pd.Index([0, 1, 2], name=name_nodes)),
    ),
    # in
    (
        tahini.core.graph.Graph(edges=[(0, 1), (0, 2)]),
        [],
        dict(column_name='x', columns_stack=[names_edges[1]]),
        pd.DataFrame(data=dict(x=[0, 1, 1]), index=pd.Index([0, 1, 2], name=name_nodes)),
    ),
    # out
    (
        tahini.core.graph.Graph(edges=[(0, 1), (0, 2)]),
        [],
        dict(column_name='x', columns_stack=[names_edges[0]]),
        pd.DataFrame(data=dict(x=[2, 0, 0]), index=pd.Index([0, 1, 2], name=name_nodes)),
    ),
])
def test_graph__edges_stack_value_counts(graph, args, kwargs, expected):
    df = graph._edges_stack_value_counts(*args, **kwargs)
    assert_frame_equal(df, expected)


@pytest.mark.parametrize('graph', [
    tahini.core.graph.Graph(),
    tahini.core.graph.UndirectedGraph(),
])
def test_graph__info_columns_degree(graph):
    info = graph._info_columns_degree()
    assert isinstance(info, dict)
    assert all([item in info for item in ['degree', 'degree_in', 'degree_out']])


@pytest.mark.parametrize('graph, expected', [
    # empty
    (
        tahini.core.graph.Graph(),
        pd.DataFrame(columns=['degree_in', 'degree_out', 'degree'], index=pd.Index([], name=name_nodes)),
    ),
    (tahini.core.graph.UndirectedGraph(), pd.DataFrame(columns=['degree'], index=pd.Index([], name=name_nodes))),
    # non empty
    (
        tahini.core.graph.Graph(edges=[(0, 1)]),
        pd.DataFrame(
            data=dict(degree_in=[0, 1], degree_out=[1, 0], degree=[1, 1]),
            index=pd.Index([0, 1], name=name_nodes),
        ),
    ),
    # non empty with zero degree
    (
        tahini.core.graph.Graph(nodes=[2], edges=[(0, 1)]),
        pd.DataFrame(
            data=dict(degree_in=[0, 1, 0], degree_out=[1, 0, 0], degree=[1, 1, 0]),
            index=pd.Index([0, 1, 2], name=name_nodes),
        ),
    ),
])
def test_graph_degrees(graph, expected):
    df = graph.degrees
    assert_frame_equal(df, expected)


@pytest.mark.parametrize('graph, expected', [
    # empty
    (
        tahini.core.graph.Graph(),
        pd.DataFrame(columns=['neighbors_in', 'neighbors_out', 'neighbors'], index=pd.Index([], name=name_nodes)),
    ),
    (tahini.core.graph.UndirectedGraph(), pd.DataFrame(columns=['neighbors'], index=pd.Index([], name=name_nodes))),
    # non empty
    (
        tahini.core.graph.Graph(edges=[(0, 1)]),
        pd.DataFrame(
            data=dict(neighbors_in=[[], [0]], neighbors_out=[[1], []], neighbors=[[1], [0]]),
            index=pd.Index([0, 1], name=name_nodes),
        ),
    ),
    (
        tahini.core.graph.Graph(nodes=[2], edges=[(0, 1)]),
        pd.DataFrame(
            data=dict(neighbors_in=[[], [0], []], neighbors_out=[[1], [], []], neighbors=[[1], [0], []]),
            index=pd.Index([0, 1, 2], name=name_nodes),
        ),
    ),
    (
        tahini.core.graph.Graph(edges=[(0, 1), (1, 0)]),
        pd.DataFrame(
            data=dict(neighbors_in=[[1], [0]], neighbors_out=[[1], [0]], neighbors=[[1], [0]]),
            index=pd.Index([0, 1], name=name_nodes),
        ),
    ),
    (
        tahini.core.graph.Graph(edges=[(0, 1), (1, 2), (0, 2)]),
        pd.DataFrame(
            data=dict(
                neighbors_in=[[], [0], [1, 0]],
                neighbors_out=[[1, 2], [2], []],
                neighbors=[[1, 2], [0, 2], [0, 1]],
            ),
            index=pd.Index([0, 1, 2], name=name_nodes),
        ),
    ),
])
def test_graph_get_neighbors(graph, expected):
    df = graph.neighbors
    assert_frame_equal(df, expected)



import pytest

import pandas as pd
import numpy as np

import tahini.plot.positions
import tahini.core.nodes


@pytest.mark.parametrize('args, kwargs, type_error, message_error', [
    # mismatch between dim and center length
    ([], dict(center=[0, 0], dim=1), ValueError, "Shape of 'center' has to be ('dim',)"),
    ([], dict(center=[0], dim=2), ValueError, "Shape of 'center' has to be ('dim',)"),
    ([], dict(center=[0]), ValueError, "Shape of 'center' has to be ('dim',)"),
])
def test__process_parameters_error(args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        tahini.plot.positions._process_parameters(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('args, kwargs, center_expected, dim_expected', [
    # empty
    ([], dict(), np.zeros(2), 2),
    # dim
    ([], dict(dim=1), np.zeros(1), 1),
    ([], dict(dim=2), np.zeros(2), 2),
    # center and dim
    ([], dict(center=[1], dim=1), np.ones(1), 1),
    # center
    ([], dict(center=[1, 1]), np.ones(2), 2),
])
def test__process_parameters(args, kwargs, center_expected, dim_expected):
    center, dim = tahini.plot.positions._process_parameters(*args, **kwargs)
    np.testing.assert_array_equal(center, center_expected)
    assert dim == dim_expected


def get_data_frame(*args, items=None, dim=None, **kwargs):
    if items is None:
        items = []
    if dim is None:
        dim = 2
    df = pd.DataFrame(*args, index=items, columns=[f'position_dim_{i}' for i in range(dim)], **kwargs).astype('float64')
    return df


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty items
    ([], dict(array=np.zeros((0, 2)), items=[], dim=2), get_data_frame(data=np.zeros((0, 2)))),
    # non empty items
    ([], dict(array=np.zeros((1, 2)), items=[0], dim=2), get_data_frame(data=np.zeros((1, 2)), items=[0])),
    ([], dict(array=np.ones((1, 2)), items=[0], dim=2), get_data_frame(data=np.ones((1, 2)), items=[0])),
    # non default dim
    ([], dict(array=np.ones((0, 0)), items=[], dim=0), get_data_frame(data=np.zeros((0, 0)), dim=0)),
    ([], dict(array=np.ones((0, 1)), items=[], dim=1), get_data_frame(data=np.zeros((0, 1)), dim=1)),
    (
        [],
        dict(array=np.zeros((1, 1)), items=[0], dim=1),
        get_data_frame(data=np.zeros((1, 1)), items=[0], dim=1),
    ),
    (
        [],
        dict(array=np.zeros((1, 3)), items=[0], dim=3),
        get_data_frame(data=np.zeros((1, 3)), items=[0], dim=3),
    ),
    (
        [],
        dict(array=np.zeros((2, 3)), items=[0, 1], dim=3),
        get_data_frame(data=np.zeros((2, 3)), items=[0, 1], dim=3),
    ),
])
def test__array_to_data_frame(args, kwargs, expected):
    df = tahini.plot.positions._array_to_data_frame(*args, **kwargs)
    pd.testing.assert_frame_equal(df, expected)


@pytest.mark.parametrize('args, kwargs, type_error, message_error', [
    ([], dict(items=[], dim=1), ValueError, "'dim' has to be > 1 for circular layout"),
])
def test__get_circular_error(args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        tahini.plot.positions._get_circular(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty items
    ([], dict(items=[], dim=2), np.zeros((0, 2))),
    # empty items non one scale
    ([], dict(items=[], dim=2, scale=2), np.zeros((0, 2))),
    # items with order 1
    ([], dict(items=[0], dim=2), np.zeros((1, 2))),
    # items with order 2
    ([], dict(items=[0, 1], dim=2), np.array([[1, 0], [-1, 0]])),
    # items with order 2 and non one scale
    ([], dict(items=[0, 1], dim=2, scale=2), np.array([[2, 0], [-2, 0]])),
    # items with order 2 and dim > 2
    ([], dict(items=[0, 1], dim=3), np.array([[1, 0, 0], [-1, 0, 0]])),
])
def test__get_circular(args, kwargs, expected):
    positions = tahini.plot.positions._get_circular(*args, **kwargs)
    np.testing.assert_array_almost_equal(positions, expected)


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty items
    ([], dict(items=[], dim=2), np.zeros((0, 2))),
    # empty items with seed
    ([], dict(items=[], dim=2, seed=0), np.zeros((0, 2))),
    # items of order 1
    ([], dict(items=[0], dim=2, seed=0), np.array([[0.548814, 0.715189]])),
    # items of order 2
    ([], dict(items=[0, 1], dim=2, seed=0), np.array([[0.548814, 0.715189], [0.602763, 0.544883]])),
    # items of order 2 and dim 3
    (
        [],
        dict(items=[0, 1], dim=3, seed=0),
        np.array([[0.548814, 0.715189, 0.602763], [0.544883, 0.423655, 0.645894]]),
    ),
])
def test__get_random(args, kwargs, expected):
    positions = tahini.plot.positions._get_random(*args, **kwargs)
    np.testing.assert_array_almost_equal(positions, expected)


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty items
    ([], dict(items=[]), get_data_frame()),
    # non default dim
    ([], dict(items=[], dim=3), get_data_frame(dim=3)),
    # items of order 1
    ([], dict(items=[0]), get_data_frame(data=[[0, 0]], items=range(1))),
    # items of order 2
    ([], dict(items=[0, 1]), get_data_frame(data=[[1, 0], [-1, 0]], items=range(2))),
    # layout - circular
    (
        [],
        dict(items=[0, 1], layout='circular'),
        get_data_frame(data=[[1, 0], [-1, 0]], items=range(2)),
    ),
    # layout - random
    (
        [],
        dict(items=[0, 1], layout='random', seed=0),
        get_data_frame(data=[[0.548814, 0.715189], [0.602763, 0.544883]], items=range(2)),
    ),
    # center
    ([], dict(items=[0], center=[1, 1]), get_data_frame(data=[[1, 1]], items=range(1))),
])
def test_get(args, kwargs, expected):
    positions = tahini.plot.positions.get(*args, **kwargs)
    pd.testing.assert_frame_equal(positions, expected)



from typing import Optional, Union, NoReturn

import pandas as pd

from tahini.core.base import ContainerDataIndexed
from tahini.core import Graph

__all__ = ['assert_graph_equal', 'assert_container_equal']


def assert_container_equal(
        left: ContainerDataIndexed,
        right: ContainerDataIndexed,
        check_dtype: Optional[bool] = None,
        check_index_type: Optional[Union[bool, str]] = None,
        check_column_type: Optional[Union[bool, str]] = None,
        check_like: Optional[bool] = None,
        obj: Optional[str] = None,
        **kwargs,
) -> NoReturn:

    if check_dtype is None:
        check_dtype = False
    if check_index_type is None:
        check_index_type = False
    if check_column_type is None:
        check_column_type = False
    if check_like is None:
        check_like = True

    assert type(left) is type(right), f"""Types are different

[left]:  {left.__class__.__name__}
[right]: {right.__class__.__name__}"""

    if obj is None:
        obj = f'{left.__class__.__name__}.data_testing'

    pd.testing.assert_frame_equal(
        left=left.data_testing,
        right=right.data_testing,
        check_dtype=check_dtype,
        check_index_type=check_index_type,
        check_column_type=check_column_type,
        check_like=check_like,
        obj=obj,
        **kwargs,
    )


def assert_graph_equal(
        left: Graph,
        right: Graph,
        obj: Optional[str] = None,
        **kwargs,
) -> NoReturn:

    if obj is not None:
        flag_obj = True
    else:
        flag_obj = False

    if not flag_obj:
        obj = f'{left.__class__.__name__}.nodes.data_testing'
    assert_container_equal(left=left.nodes, right=right.nodes, obj=obj, **kwargs)

    if not flag_obj:
        obj = f'{left.__class__.__name__}.edges.data_testing'
    assert_container_equal(left=left.edges, right=right.edges, obj=obj, **kwargs)



from __future__ import annotations
from typing import Optional, Union, Sequence

from pandas import DataFrame

from .base import (
    ContainerDataIndexed,
    TypeIndexInput,
    TypeDataInput,
)
from ..plot.positions import get as get_positions

__all__ = [
    'Nodes',
    'TypeNodesInput',
]


class Nodes(ContainerDataIndexed):

    _names_index = ['node']
    _name_index_internal = 'node_internal'

    def __init__(
            self,
            index: Optional[TypeIndexInput] = None,
            data: Optional[TypeDataInput] = None,
            order: Optional[int] = None,
            **kwargs,
    ):

        if order is not None:
            index = range(order)

        super().__init__(index=index, data=data, **kwargs)

    def get_positions(
            self,
            layout: Optional[str] = None,
            center: Optional[Sequence] = None,
            dim: Optional[int] = None,
            **kwargs,
    ) -> DataFrame:

        df = Nodes(data=get_positions(
            items=self.data.index,
            layout=layout,
            center=center,
            dim=dim,
            **kwargs,
        )).data

        return df


TypeNodesInput = Union[Nodes, TypeIndexInput]



from typing import Optional, Sequence, Tuple

import pandas as pd
import numpy as np

__all__ = ['get']


# All credit for this goes to networkx https://github.com/networkx/networkx/blob/master/networkx/drawing/layout.py


def _process_parameters(
        center: Optional[Sequence] = None,
        dim: Optional[int] = None,
) -> Tuple[np.array, int]:

    if dim is None:
        dim = 2

    if center is None:
        center = np.zeros(dim)
    else:
        center = np.asarray(center)

    if center.shape != (dim,):
        raise ValueError("Shape of 'center' has to be ('dim',)")

    return center, dim


def _array_to_data_frame(
        array: np.array,
        items: Sequence,
        dim: int,
) -> pd.DataFrame:

    df = pd.DataFrame(
        data=array,
        index=items,
        columns=[f'position_dim_{i}' for i in range(dim)],
    ).astype('float64')

    return df


def get(
        items: Sequence,
        layout: Optional[str] = None,
        center: Optional[Sequence] = None,
        dim: Optional[int] = None,
        **kwargs,
) -> pd.DataFrame:

    if layout is None:
        layout = 'circular'

    center, dim = _process_parameters(center=center, dim=dim)

    function_get_positions_array = globals()[f'_get_{layout}']
    positions_base = function_get_positions_array(items=items, dim=dim, **kwargs)

    positions = positions_base + center

    df = _array_to_data_frame(array=positions, items=items, dim=dim)

    return df


def _get_circular(
        items: Sequence,
        dim: int,
        scale: Optional[int] = None,
) -> np.array:

    if scale is None:
        scale = 1

    if dim < 2:
        raise ValueError("'dim' has to be > 1 for circular layout")

    pad_dims = max(0, dim - 2)

    length_items = len(items)

    if length_items == 1:
        positions = np.zeros((1, dim))
    else:
        theta = np.linspace(start=0, stop=1, num=length_items, endpoint=False) * 2 * np.pi
        positions = np.column_stack(
            [np.cos(theta), np.sin(theta), np.zeros((length_items, pad_dims))]
        )

    positions = positions * scale

    return positions


def _get_random(
        items: Sequence,
        dim: int,
        seed: Optional[int] = None,
) -> np.array:
    rs = np.random.RandomState(seed=seed)
    positions = rs.rand(len(items), dim)
    return positions



from collections.abc import Sequence
from functools import partial
from math import isnan, nan

import pytest
from hypothesis import given
import hypothesis.strategies as st
from hypothesis.extra.pandas import indexes, columns, data_frames
import pandas as pd

import tahini.core.base
import tahini.testing

names_index_container_data_indexed = 'index'
name_index_internal = 'index_internal'
names_index_container_data_indexed_multi = ('index_0', 'index_1')


def get_data_frame(*args, name_index=names_index_container_data_indexed, **kwargs) -> pd.DataFrame:
    return pd.DataFrame(*args, **kwargs).rename_axis(index=name_index)


def get_data_frame_internal(
        *args,
        index_internal=None,
        name_index=names_index_container_data_indexed,
        **kwargs,
) -> pd.DataFrame:
    df = pd.DataFrame(*args, **kwargs).rename_axis(index=name_index).reset_index()
    if index_internal is None:
        index_internal = df[name_index]
    df.index = pd.Index(index_internal, name=name_index_internal)
    return df


def get_data_frame_index_multi(
        *args,
        names_index=names_index_container_data_indexed_multi,
        index=None,
        **kwargs,
) -> pd.DataFrame:
    if index is None:
        index = pd.MultiIndex(levels=[[]] * len(names_index), codes=[[]] * len(names_index), names=names_index)
    else:
        index = pd.MultiIndex.from_tuples(index, names=names_index)
    return pd.DataFrame(*args, index=index, **kwargs)


def get_data_frame_internal_index_multi(
        *args,
        index_internal=None,
        mapper=None,
        **kwargs,
) -> pd.DataFrame:

    df = get_data_frame_index_multi(*args, **kwargs)

    if mapper is None:
        def identity(x): return x
        mapper = identity

    if index_internal is None:
        index_internal = df.index.to_flat_index().map(mapper)

    df = df.reset_index()
    df.index = pd.Index(index_internal, name=name_index_internal)

    return df


def get_data_frame_internal_simple_index_multi(*arg, **kwargs):
    df = (
        get_data_frame_internal_index_multi(*arg, **kwargs)
        .drop(columns=list(names_index_container_data_indexed_multi))
    )
    return df


get_data_frame_internal_index_multi_sets = partial(get_data_frame_internal_index_multi, mapper=frozenset)

get_data_frame_internal_simple_index_multi_sets = partial(
    get_data_frame_internal_simple_index_multi,
    mapper=frozenset,
)

assert_frame_equal = partial(
    pd.testing.assert_frame_equal,
    check_dtype=False,
    check_column_type=False,
    check_index_type=False,
)

assert_index_equal = partial(pd.testing.assert_index_equal, exact=False)


def check_nan(x):
    try:
        tf = isnan(x)
    except TypeError:
        tf = False
    return tf


@pytest.mark.parametrize('klass', [
    tahini.core.base.ContainerDataIndexed,
    tahini.core.base.ContainerDataIndexedMulti,
    tahini.core.base.ContainerDataIndexedMultiSets,
])
def test_container_data_indexed__names_index(klass):
    assert isinstance(klass._names_index, Sequence)


@pytest.mark.parametrize('klass', [
    tahini.core.base.ContainerDataIndexed,
    tahini.core.base.ContainerDataIndexedMulti,
    tahini.core.base.ContainerDataIndexedMultiSets,
])
def test_container_data_indexed__name_index_internal(klass):
    assert isinstance(klass._name_index_internal, str)


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty index
    ([], dict(index=pd.Index([])), pd.Index([])),
    # non empty index
    ([], dict(index=pd.Index([0])), pd.Index([0])),
    # empty multi index
    ([], dict(index=pd.MultiIndex.from_arrays([[]])), pd.MultiIndex.from_arrays([[]])),
])
def test_container_data_indexed__create_index_internal(args, kwargs, expected):
    index = tahini.core.base.ContainerDataIndexed._create_index_internal(*args, **kwargs)
    assert_index_equal(index, expected)


@pytest.mark.parametrize('args, kwargs, type_error, message_error', [
    # non unique index
    ([], dict(index=pd.Index([0, 0])), ValueError, "Index needs to be unique for 'ContainerDataIndexed'"),
])
def test_container_data_indexed__validate_index_error(args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        tahini.core.base.ContainerDataIndexed._validate_index(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty
    ([], dict(), get_data_frame_internal()),
    # non empty index
    ([], dict(index=[0]), get_data_frame_internal(index=[0])),
    # empty index
    ([], dict(index=[]), get_data_frame_internal()),
    # empty container idempotent
    ([], dict(index=tahini.core.base.ContainerDataIndexed()), get_data_frame_internal()),
    # empty data dict
    ([], dict(data=dict()), get_data_frame_internal()),
    # empty data records
    ([], dict(data=[]), get_data_frame_internal()),
    # empty data frame
    ([], dict(data=pd.DataFrame()), get_data_frame_internal()),
    # data dict
    ([], dict(data=dict(a=[1])), get_data_frame_internal(data=dict(a=[1]))),
    # dict and index
    ([], dict(data=dict(a=[1]), index=['z']), get_data_frame_internal(data=dict(a=[1]), index=['z'])),
    # data frame
    ([], dict(data=pd.DataFrame(data=dict(a=[1]))), get_data_frame_internal(data=dict(a=[1]))),
    # data frame with index
    (
        [],
        dict(data=pd.DataFrame(data=dict(a=[1]), index=['z'])),
        get_data_frame_internal(data=dict(a=[1]), index=['z']),
    ),
    # data frame and index
    (
        [],
        dict(data=pd.DataFrame(data=dict(a=[1])), index=['z']),
        get_data_frame_internal(data=dict(a=[1]), index=['z']),
    ),
    # data records
    ([], dict(data=[[1]]), get_data_frame_internal(data=[[1]])),
    ([], dict(data=['a', 'b']), get_data_frame_internal({0: ['a', 'b']})),
    ([], dict(data=[['a'], ['b']]), get_data_frame_internal({0: ['a', 'b']})),
    ([], dict(data=[['a', 'b']]), get_data_frame_internal({0: ['a'], 1: ['b']})),
    # container idempotent
    (
        [],
        dict(index=tahini.core.base.ContainerDataIndexed(data=pd.DataFrame(data=dict(a=[1]), index=['z']))),
        get_data_frame_internal(data=dict(a=[1]), index=['z']),
    ),
    # index as column
    ([], dict(data=dict(index=[0, 1])), get_data_frame_internal(index=[0, 1])),
])
def test_container_data_indexed_init(args, kwargs, expected):
    container = tahini.core.base.ContainerDataIndexed(*args, **kwargs)
    assert_frame_equal(container.data_internal, expected)


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty
    ([], dict(index=pd.Index([])), pd.Index([], name=names_index_container_data_indexed)),
    # non empty
    ([], dict(index=pd.Index([0])), pd.Index([0], name=names_index_container_data_indexed)),
])
def test_container_data_indexed__validate_index(args, kwargs, expected):
    index = tahini.core.base.ContainerDataIndexed._validate_index(*args, **kwargs)
    assert_index_equal(index, expected)


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty
    ([], dict(data=pd.DataFrame()), get_data_frame()),
    # non empty index
    ([], dict(data=pd.DataFrame(index=['a', 'b'])), get_data_frame(index=['a', 'b'])),
    # non empty index with name
    (
        [],
        dict(data=pd.DataFrame(index=pd.Index(['a', 'b'], name=f'not_{names_index_container_data_indexed}'))),
        get_data_frame(index=['a', 'b']),
    ),
    # non empty data
    ([], dict(data=pd.DataFrame(data=dict(a=[0, 1], b=[0, 1]))), get_data_frame(data=dict(a=[0, 1], b=[0, 1]))),
])
def test_container_data_indexed__validate_data(args, kwargs, expected):
    df = tahini.core.base.ContainerDataIndexed._validate_data(*args, **kwargs)
    assert_frame_equal(df, expected)


@pytest.mark.parametrize('args, kwargs, type_error, message_error', [
    # non unique index
    ([], dict(index=[0, 0]), ValueError, "Index needs to be unique for 'ContainerDataIndexed'"),
    # non matching length between index and data
    (
        [],
        dict(data=pd.DataFrame(data=dict(a=[1])), index=[0, 1]),
        ValueError,
        "Length mismatch: Expected axis has 1 elements, new values have 2 elements",
    ),
    # non matching length between index and data
    (
        [],
        dict(data=pd.DataFrame(data=dict(a=[1, 2])), index=[0]),
        ValueError,
        "Length mismatch: Expected axis has 2 elements, new values have 1 elements",
    ),
])
def test_container_data_indexed_init_error(args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        tahini.core.base.ContainerDataIndexed(*args, **kwargs)
    assert e.value.args[0] == message_error


types_index = (
    st.iterables,
    indexes,
)

elements_non_specific = (
    st.binary,
    st.booleans,
    st.characters,
    st.complex_numbers,
    st.dates,
    st.datetimes,
    st.fractions,
    st.integers,
    st.none,
    st.randoms,
    st.text,
    st.times,
    st.uuids,
)

elements_specific = (
    # pandas.Timedeltas max and min do not match python standard library datetime.timedelta max and min
    (
        st.timedeltas,
        dict(min_value=pd.Timedelta.min.to_pytimedelta(), max_value=pd.Timedelta.max.to_pytimedelta()),
        lambda x: True,
        lambda x: True,
    ),
    # error with decimals not being able to hash snan
    (
        st.decimals,
        dict(),
        lambda x: not x.is_snan(),
        lambda x: True,
    ),
    # cannot have duplicate nans
    (
        st.floats,
        dict(),
        lambda x: True,
        lambda container: sum([isnan(item) for item in container]) < 2,
    ),
)

# todo fix https://github.com/tahini-dev/tahini/runs/1898415631?check_suite_focus=true


@pytest.mark.parametrize('type_index', types_index)
@pytest.mark.parametrize('elements, kwargs_elements, filter_elements, filter_type_index', [
    *((item, dict(), lambda x: True, lambda x: True) for item in elements_non_specific),
    *elements_specific,
])
@given(data=st.data())
def test_container_data_indexed_init_index_single_elements_type(
        type_index,
        elements,
        kwargs_elements,
        filter_elements,
        filter_type_index,
        data,
):
    container = tahini.core.base.ContainerDataIndexed(
        index=data.draw(
            type_index(
                elements=elements(**kwargs_elements).filter(filter_elements),
                unique=True,
            )
            .filter(filter_type_index)
        ),
    )
    assert isinstance(container.data_internal, pd.DataFrame)


@pytest.mark.parametrize('type_index', types_index)
@pytest.mark.parametrize('elements', [
    pytest.param(
        st.timedeltas,
        marks=pytest.mark.xfail(
            reason='pandas.Timedeltas max and min do not match python standard library datetime.timedelta max and min',
        ),
    ),
    pytest.param(
        st.decimals,
        marks=pytest.mark.xfail(
            reason='error with decimals not being able to hash snan',
        ),
    ),
    pytest.param(
        st.floats,
        marks=pytest.mark.xfail(
            reason='error with duplicate nans',
        ),
    ),
])
@given(data=st.data())
def test_container_data_indexed_init_index_single_elements_type_x_fail(type_index, elements, data):
    container = tahini.core.base.ContainerDataIndexed(
        index=data.draw(type_index(elements=elements.param(), unique=True)),
    )


@pytest.fixture(scope='module')
def list_elements():
    output_value = (
        *(elements() for elements in elements_non_specific),
        *(item[0](**item[1]).filter(item[2]) for item in elements_specific),
    )
    return output_value


@pytest.mark.parametrize('type_index', types_index)
@given(data=st.data())
def test_container_data_indexed_init_index_multiple_elements_type(type_index, list_elements, data):
    index = data.draw(
        type_index(elements=st.one_of(*list_elements), unique=True)
        .filter(lambda x: sum([check_nan(item) for item in x]) < 2)
    )
    container = tahini.core.base.ContainerDataIndexed(index=index)
    assert isinstance(container.data_internal, pd.DataFrame)


@given(data=data_frames(columns=(columns('A', elements=st.integers())), index=indexes(elements=st.integers())))
def test_container_data_indexed_init_data_data_frame(data):
    container = tahini.core.base.ContainerDataIndexed(data=data)
    assert isinstance(container.data_internal, pd.DataFrame)


@pytest.mark.parametrize('container, expected', [
    (tahini.core.base.ContainerDataIndexed(), [names_index_container_data_indexed]),
    (tahini.core.base.ContainerDataIndexedMulti(), list(names_index_container_data_indexed_multi)),
    (tahini.core.base.ContainerDataIndexedMultiSets(), list(names_index_container_data_indexed_multi)),
])
def test_container_data_indexed_names_index(container, expected):
    names_index = container.names_index
    assert names_index == expected


@pytest.mark.parametrize('container, data, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexed(), pd.DataFrame(), get_data_frame()),
    # non empty container
    (tahini.core.base.ContainerDataIndexed(data=dict(a=['1'])), pd.DataFrame(), get_data_frame()),
    # empty container and non empty data
    (
        tahini.core.base.ContainerDataIndexed(),
        pd.DataFrame(data=dict(a=['1'])),
        get_data_frame(data=dict(a=['1'])),
    ),
    # non empty container and data
    (
        tahini.core.base.ContainerDataIndexed(data=dict(a=['1'])),
        pd.DataFrame(data=dict(b=[2])),
        get_data_frame(data=dict(b=[2])),
    ),
])
def test_container_data_indexed_data(container, data, expected):
    container.data = data
    assert_frame_equal(container.data, expected)


@pytest.mark.parametrize('container, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexed(), get_data_frame_internal()),
    # non empty container
    (
        tahini.core.base.ContainerDataIndexed(data=dict(a=['1'])),
        get_data_frame_internal(data=dict(a=['1'])),
    ),
])
def test_container_data_indexed_data_internal(container, expected):
    assert_frame_equal(container.data_internal, expected)


@pytest.mark.parametrize('container, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexed(), get_data_frame(name_index=name_index_internal)),
    # non empty container
    (
        tahini.core.base.ContainerDataIndexed(data=dict(a=['1'])),
        get_data_frame(data=dict(a=['1']), name_index=name_index_internal),
    ),
])
def test_container_data_indexed_data_testing(container, expected):
    assert_frame_equal(container.data_testing, expected)


@pytest.mark.parametrize('container, args, kwargs, type_error, message_error', [
    # index not in container
    (tahini.core.base.ContainerDataIndexed(), [], dict(index=[0]), KeyError, "[0] not found in axis")
])
def test_container_data_indexed_drop_error(container, args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        container.drop(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('container, args, kwargs, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexed(), [], dict(), tahini.core.base.ContainerDataIndexed()),
    # empty container ignore error
    (
        tahini.core.base.ContainerDataIndexed(),
        [],
        dict(index=[0], errors='ignore'),
        tahini.core.base.ContainerDataIndexed(),
    ),
    # empty inputs
    (
        tahini.core.base.ContainerDataIndexed(index=[0]),
        [],
        dict(),
        tahini.core.base.ContainerDataIndexed(index=[0]),
    ),
    # non empty
    (
        tahini.core.base.ContainerDataIndexed(index=[0]),
        [],
        dict(index=[0]),
        tahini.core.base.ContainerDataIndexed(),
    ),
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1]),
        [],
        dict(index=[0, 1]),
        tahini.core.base.ContainerDataIndexed(),
    ),
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1]),
        [],
        dict(index=[1]),
        tahini.core.base.ContainerDataIndexed(index=[0]),
    ),
    # drop columns
    (
        tahini.core.base.ContainerDataIndexed(index=[0], columns=['a']),
        [],
        dict(columns=['a']),
        tahini.core.base.ContainerDataIndexed(index=[0]),
    ),
])
def test_container_data_indexed_drop(container, args, kwargs, expected):
    container = container.drop(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(container, expected)


@pytest.mark.parametrize('container, args, kwargs, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexed(), [], dict(), tahini.core.base.ContainerDataIndexed()),
    # empty column
    (
        tahini.core.base.ContainerDataIndexed(),
        [],
        dict(data=pd.DataFrame(data=dict(value=[]))),
        tahini.core.base.ContainerDataIndexed(data=dict(value=[])),
    ),
    # empty inputs
    (
        tahini.core.base.ContainerDataIndexed(index=[0]),
        [],
        dict(),
        tahini.core.base.ContainerDataIndexed(index=[0]),
    ),
    # empty container and non empty index
    (
        tahini.core.base.ContainerDataIndexed(),
        [],
        dict(index=[0]),
        tahini.core.base.ContainerDataIndexed(index=[0]),
    ),
    # empty container and non empty data
    (
        tahini.core.base.ContainerDataIndexed(),
        [],
        dict(data=dict(a=[1])),
        tahini.core.base.ContainerDataIndexed(data=dict(a=[1])),
    ),
    # update with no new changes
    (
        tahini.core.base.ContainerDataIndexed(index=[0]),
        [],
        dict(index=[0]),
        tahini.core.base.ContainerDataIndexed(index=[0]),
    ),
    # update seems to sort
    (
        tahini.core.base.ContainerDataIndexed(index=[0]),
        [],
        dict(index=[2, 1]),
        tahini.core.base.ContainerDataIndexed(index=[0, 1, 2]),
    ),
    # new column and index
    (
        tahini.core.base.ContainerDataIndexed(),
        [],
        dict(data=dict(a=[1, 2])),
        tahini.core.base.ContainerDataIndexed(data=dict(a=[1, 2])),
    ),
    # new column for given index
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1]),
        [],
        dict(index=[0, 1], data=dict(a=[1, 2])),
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[1, 2])),
    ),
    # new column and index item
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1]),
        [],
        dict(index=[1, 2], data=dict(a=[1, 2])),
        tahini.core.base.ContainerDataIndexed(index=[0, 1, 2], data=dict(a=[nan, 1, 2])),
    ),
    # cannot update to nan with default func
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[1, 2])),
        [],
        dict(index=[1], data=dict(a=[nan])),
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[1, 2])),
    ),
    # single value in column
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[1, 2])),
        [],
        dict(index=[1], data=dict(a=[3])),
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[1, 3])),
    ),
    # single value in column
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[1, 2])),
        [],
        dict(index=[0], data=dict(a=[3])),
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[3, 2])),
    ),
    # new additional column
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[1, 2])),
        [],
        dict(index=[0, 1], data=dict(b=[2, 3])),
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[1, 2], b=[2, 3])),
    ),
    # row update
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[1, 2], b=[2, 3])),
        [],
        dict(index=[0], data=dict(a=[4], b=[5])),
        tahini.core.base.ContainerDataIndexed(index=[0, 1], data=dict(a=[4, 2], b=[5, 3])),
    ),
])
def test_container_data_indexed_update(container, args, kwargs, expected):
    container = container.update(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(container, expected)


@pytest.mark.parametrize('container, args, kwargs, type_error, message_error', [
    # missing map multiple
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1]),
        [],
        dict(mapper={}),
        ValueError,
        "Index needs to be unique for 'ContainerDataIndexed'",
    ),
])
def test_container_data_indexed_map_error(container, args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        container.map(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('container, args, kwargs, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexed(), [], dict(), tahini.core.base.ContainerDataIndexed()),
    # empty inputs
    (
        tahini.core.base.ContainerDataIndexed(index=[0]),
        [],
        dict(),
        tahini.core.base.ContainerDataIndexed(index=[0]),
    ),
    # empty container
    (
        tahini.core.base.ContainerDataIndexed(),
        [],
        dict(mapper=dict()),
        tahini.core.base.ContainerDataIndexed(),
    ),
    (
        tahini.core.base.ContainerDataIndexed(),
        [],
        dict(mapper=dict(a=1)),
        tahini.core.base.ContainerDataIndexed(),
    ),
    # non empty
    (
        tahini.core.base.ContainerDataIndexed(index=[0]),
        [],
        dict(mapper={0: 1}), tahini.core.base.ContainerDataIndexed(index=[1]),
    ),
    # change index type
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1]),
        [],
        dict(mapper={0: 'a', 1: 'b'}),
        tahini.core.base.ContainerDataIndexed(index=['a', 'b']),
    ),
    # missing map
    (
        tahini.core.base.ContainerDataIndexed(index=[0, 1]),
        [],
        dict(mapper={0: 'a'}),
        tahini.core.base.ContainerDataIndexed(index=['a', nan]),
    ),
])
def test_container_data_indexed_map(container, args, kwargs, expected):
    container = container.map(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(container, expected)


@pytest.mark.parametrize('container, expected', [
    (tahini.core.base.ContainerDataIndexed(), f'ContainerDataIndexed(index={get_data_frame().index})'),
    (
        tahini.core.base.ContainerDataIndexedMulti(),
        f'ContainerDataIndexedMulti(index={get_data_frame_index_multi().index})',
    ),
    (
        tahini.core.base.ContainerDataIndexedMultiSets(),
        f'ContainerDataIndexedMultiSets(index={get_data_frame_index_multi().index})',
    ),
])
def test_container_data_indexed_repr(container, expected):
    repr_container = repr(container)
    assert repr_container == expected


@pytest.mark.parametrize('container, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexed(), []),
    (tahini.core.base.ContainerDataIndexedMulti(), []),
    (tahini.core.base.ContainerDataIndexedMultiSets(), []),
    # non empty
    (tahini.core.base.ContainerDataIndexed(index=[0]), [0]),
    (tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1), (0, 2)]), [(0, 1), (0, 2)]),
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1), (0, 2)]),
        [frozenset((0, 1)), frozenset((0, 2))],
    ),
])
def test_container_data_indexed_iter(container, expected):
    assert [item for item in container.iter()] == expected
    assert [item for item in container] == expected


@pytest.mark.parametrize('container, item, expected', [
    # not in empty
    (tahini.core.base.ContainerDataIndexed(), 0, False),
    (tahini.core.base.ContainerDataIndexedMulti(), (0, 1), False),
    (tahini.core.base.ContainerDataIndexedMultiSets(), (0, 1), False),
    # contains
    (tahini.core.base.ContainerDataIndexed(index=[0]), 0, True),
    (tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]), (0, 1), True),
    (tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]), (0, 1), True),
    (tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]), (1, 0), True),
    # not contains
    (tahini.core.base.ContainerDataIndexed(index=[0]), 1, False),
    (tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]), (0, 2), False),
    (tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]), (1, 0), False),
    (tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]), (0, 2), False),
])
def test_container_data_indexed_contains(container, item, expected):
    assert (item in container) == expected


@pytest.mark.parametrize('container, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexed(), 0),
    (tahini.core.base.ContainerDataIndexedMulti(), 0),
    (tahini.core.base.ContainerDataIndexedMultiSets(), 0),
    # non empty
    (tahini.core.base.ContainerDataIndexed(index=[0]), 1),
    (tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]), 1),
    (tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]), 1),
])
def test_container_data_indexed_len(container, expected):
    assert len(container) == expected


@pytest.mark.parametrize('container_left, container_right, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexed(), tahini.core.base.ContainerDataIndexed(), True),
    (tahini.core.base.ContainerDataIndexedMulti(), tahini.core.base.ContainerDataIndexedMulti(), True),
    (
        tahini.core.base.ContainerDataIndexedMultiSets(),
        tahini.core.base.ContainerDataIndexedMultiSets(),
        True,
    ),
    # non empty
    (
        tahini.core.base.ContainerDataIndexed(index=[0]),
        tahini.core.base.ContainerDataIndexed(index=[0]),
        True,
    ),
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
        True,
    ),
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        True,
    ),
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(1, 0)]),
        True,
    ),
    # empty versus non empty
    (tahini.core.base.ContainerDataIndexed(), tahini.core.base.ContainerDataIndexed(index=[0]), False),
    # None right
    (tahini.core.base.ContainerDataIndexed(), None, False),
    # None left
    (None, tahini.core.base.ContainerDataIndexed(), False),
    # different order
    (
        tahini.core.base.ContainerDataIndexed(index=[1, 2]),
        tahini.core.base.ContainerDataIndexed(index=[2, 1]),
        True,
    ),
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1), (0, 2)]),
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 2), (0, 1)]),
        True,
    ),
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1), (0, 2)]),
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 2), (0, 1)]),
        True,
    ),
])
def test_container_data_indexed_eq(container_left, container_right, expected):
    assert (container_left == container_right) == expected


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty multi index
    ([], dict(index=pd.MultiIndex.from_arrays([[]])), pd.Index([])),
    # non empty multi index
    (
        [],
        dict(index=pd.MultiIndex.from_tuples([(0, 1)])),
        pd.Index([(0, 1)]).to_flat_index(),
    ),
])
def test_container_data_indexed_multi__create_index_internal(args, kwargs, expected):
    index = tahini.core.base.ContainerDataIndexedMulti._create_index_internal(*args, **kwargs)
    assert_index_equal(index, expected)


@pytest.mark.parametrize('args, kwargs, type_error, message_error', [
    # non unique index
    (
        [],
        dict(index=pd.Index([(0, 1), (0, 1)])),
        ValueError,
        "Index needs to be unique for 'ContainerDataIndexedMulti'",
    ),
])
def test_container_data_indexed_multi__validate_index_error(args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        tahini.core.base.ContainerDataIndexedMulti._validate_index(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty
    (
        [],
        dict(index=pd.Index([])),
        pd.MultiIndex(levels=[[]] * 2, codes=[[]] * 2, names=names_index_container_data_indexed_multi),
    ),
    # index
    (
        [],
        dict(index=pd.Index([(0, 1)])),
        pd.MultiIndex.from_tuples([(0, 1)], names=names_index_container_data_indexed_multi),
    ),
    # index
    (
        [],
        dict(index=pd.MultiIndex.from_tuples([(0, 1)])),
        pd.MultiIndex.from_tuples([(0, 1)], names=names_index_container_data_indexed_multi),
    ),
])
def test_container_data_indexed_multi__validate_index(args, kwargs, expected):
    index = tahini.core.base.ContainerDataIndexedMulti._validate_index(*args, **kwargs)
    assert_index_equal(index, expected)


@pytest.mark.parametrize('args, kwargs, type_error, message_error', [
    # non unique
    ([], dict(index=[(0, 1), (0, 1)]), ValueError, "Index needs to be unique for 'ContainerDataIndexedMulti'"),
    # flat index input
    ([], dict(index=[0, 1]), TypeError, "object of type 'int' has no len()"),
    # flat index input
    ([], dict(index=[0, 1, 2]), TypeError, "object of type 'int' has no len()"),
    # flat index input
    ([], dict(index=['a', 'b']), ValueError, "Length of names must match number of levels in MultiIndex."),
])
def test_container_data_indexed_multi_init_error(args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        tahini.core.base.ContainerDataIndexedMulti(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty
    ([], dict(), get_data_frame_internal_index_multi()),
    # non empty
    ([], dict(index=[(0, 1)]), get_data_frame_internal_index_multi(index=[(0, 1)])),
    # list of lists
    ([], dict(index=[[0, 1]]), get_data_frame_internal_index_multi(index=[(0, 1)])),
    # non empty
    ([], dict(index=[(0, 1), (0, 2)]), get_data_frame_internal_index_multi(index=[(0, 1), (0, 2)])),
    # order matters
    ([], dict(index=[(0, 1), (1, 0)]), get_data_frame_internal_index_multi(index=[(0, 1), (1, 0)])),
])
def test_container_data_indexed_multi_init(args, kwargs, expected):
    container = tahini.core.base.ContainerDataIndexedMulti(*args, **kwargs)
    assert_frame_equal(container.data_internal, expected)


@pytest.mark.parametrize('container, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexedMulti(), get_data_frame_index_multi()),
    # non empty
    (tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]), get_data_frame_index_multi(index=[(0, 1)])),
    # column
    (
        tahini.core.base.ContainerDataIndexedMulti(data=dict(a=[1]), index=[(0, 1)]),
        get_data_frame_index_multi(data=dict(a=[1]), index=[(0, 1)]),
    ),
])
def test_container_data_indexed_multi_data(container, expected):
    assert_frame_equal(container.data, expected)


@pytest.mark.parametrize('container, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexedMulti(), get_data_frame_internal_index_multi()),
    # non empty
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
        get_data_frame_internal_index_multi(index=[(0, 1)]),
    ),
    # column
    (
        tahini.core.base.ContainerDataIndexedMulti(data=dict(a=[1]), index=[(0, 1)]),
        get_data_frame_internal_index_multi(data=dict(a=[1]), index=[(0, 1)]),
    ),
])
def test_container_data_indexed_multi_data_internal(container, expected):
    assert_frame_equal(container.data_internal, expected)


@pytest.mark.parametrize('container, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexedMulti(), get_data_frame_internal_simple_index_multi()),
    # non empty
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
        get_data_frame_internal_simple_index_multi(index=[(0, 1)]),
    ),
    # column
    (
        tahini.core.base.ContainerDataIndexedMulti(data=dict(a=[1]), index=[(0, 1)]),
        get_data_frame_internal_simple_index_multi(data=dict(a=[1]), index=[(0, 1)]),
    ),
])
def test_container_data_indexed_multi_data_testing(container, expected):
    assert_frame_equal(container.data_testing, expected)


@pytest.mark.parametrize('container, args, kwargs, type_error, message_error', [
    # index flat
    (
        tahini.core.base.ContainerDataIndexedMulti(),
        [],
        dict(index=[0]),
        TypeError,
        "object of type 'int' has no len()",
    ),
    # index not exist
    (
        tahini.core.base.ContainerDataIndexedMulti(),
        [],
        dict(index=[(0, 1)]),
        KeyError,
        "[(0, 1)] not found in axis",
    ),
])
def test_container_data_indexed_multi_drop_error(container, args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        container.drop(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('container, args, kwargs, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexedMulti(), [], dict(), tahini.core.base.ContainerDataIndexedMulti()),
    # empty container ignore error
    (
        tahini.core.base.ContainerDataIndexedMulti(),
        [],
        dict(index=[(0, 1)], errors='ignore'),
        tahini.core.base.ContainerDataIndexedMulti(),
    ),
    # empty inputs
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
        [],
        dict(),
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
    ),
    # non empty
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
        [],
        dict(index=[(0, 1)]),
        tahini.core.base.ContainerDataIndexedMulti(),
    ),
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1), (0, 2)]),
        [],
        dict(index=[(0, 1), (0, 2)]),
        tahini.core.base.ContainerDataIndexedMulti(),
    ),
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1), (0, 2)]),
        [],
        dict(index=[(0, 2)]),
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
    ),
    # order matters
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
        [],
        dict(index=[(1, 0)], errors='ignore'),
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
    ),
    # drop columns
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)], columns=['a']),
        [],
        dict(columns=['a']),
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
    ),
])
def test_container_data_indexed_multi_drop(container, args, kwargs, expected):
    container = container.drop(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(container, expected)


@pytest.mark.parametrize('container, args, kwargs, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexedMulti(), [], dict(), tahini.core.base.ContainerDataIndexedMulti()),
    # empty inputs
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
        [],
        dict(),
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
    ),
    # non empty
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
        [],
        dict(index=[(0, 2)]),
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1), (0, 2)]),
    ),
    # order matters
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1)]),
        [],
        dict(index=[(1, 0)]),
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1), (1, 0)]),
    ),
])
def test_container_data_indexed_multi_update(container, args, kwargs, expected):
    container = container.update(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(container, expected)


@pytest.mark.parametrize('container, args, kwargs, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexedMulti(), [], dict(), tahini.core.base.ContainerDataIndexedMulti()),
    # non empty
    (
        tahini.core.base.ContainerDataIndexedMulti(index=[(0, 1), (0, 2)]),
        [],
        dict(mapper={0: 'a', 1: 'b', 2: 'c'}),
        tahini.core.base.ContainerDataIndexedMulti(index=[('a', 'b'), ('a', 'c')]),
    ),
])
def test_container_data_indexed_multi_map(container, args, kwargs, expected):
    container = container.map(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(container, expected)


@pytest.mark.parametrize('args, kwargs, type_error, message_error', [
    # non unique
    ([], dict(index=[(0, 1), (0, 1)]), ValueError, "Index needs to be unique for 'ContainerDataIndexedMultiSets'"),
    # non unique
    ([], dict(index=[(0, 1), (1, 0)]), ValueError, "Index needs to be unique for 'ContainerDataIndexedMultiSets'"),
    # flat index input
    ([], dict(index=[0, 1]), TypeError, "object of type 'int' has no len()"),
    # flat index input
    ([], dict(index=['a', 'b']), ValueError, "Length of names must match number of levels in MultiIndex."),
])
def test_container_data_indexed_multi_sets_init_error(args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        tahini.core.base.ContainerDataIndexedMultiSets(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty
    ([], dict(), get_data_frame_internal_index_multi_sets()),
    # non empty
    ([], dict(index=[(0, 1)]), get_data_frame_internal_index_multi_sets(index=[(0, 1)])),
    # non empty
    ([], dict(index=[(0, 1), (0, 2)]), get_data_frame_internal_index_multi_sets(index=[(0, 1), (0, 2)])),
])
def test_container_data_indexed_multi_sets_init(args, kwargs, expected):
    container = tahini.core.base.ContainerDataIndexedMultiSets(*args, **kwargs)
    assert_frame_equal(container.data_internal, expected)


@pytest.mark.parametrize('container, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexedMultiSets(), get_data_frame_index_multi()),
    # non empty
    (tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]), get_data_frame_index_multi(index=[(0, 1)])),
    # column
    (
        tahini.core.base.ContainerDataIndexedMultiSets(data=dict(a=[1]), index=[(0, 1)]),
        get_data_frame_index_multi(data=dict(a=[1]), index=[(0, 1)]),
    ),
])
def test_container_data_indexed_multi_sets_data(container, expected):
    assert_frame_equal(container.data, expected)


@pytest.mark.parametrize('container, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexedMultiSets(), get_data_frame_internal_index_multi_sets()),
    # non empty
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        get_data_frame_internal_index_multi_sets(index=[(0, 1)]),
    ),
    # column
    (
        tahini.core.base.ContainerDataIndexedMultiSets(data=dict(a=[1]), index=[(0, 1)]),
        get_data_frame_internal_index_multi_sets(data=dict(a=[1]), index=[(0, 1)]),
    ),
])
def test_container_data_indexed_multi_sets_data_internal(container, expected):
    assert_frame_equal(container.data_internal, expected)


@pytest.mark.parametrize('container, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexedMultiSets(), get_data_frame_internal_simple_index_multi_sets()),
    # non empty
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        get_data_frame_internal_simple_index_multi_sets(index=[(0, 1)]),
    ),
    # column
    (
        tahini.core.base.ContainerDataIndexedMultiSets(data=dict(a=[1]), index=[(0, 1)]),
        get_data_frame_internal_simple_index_multi_sets(data=dict(a=[1]), index=[(0, 1)]),
    ),
])
def test_container_data_indexed_multi_sets_data_testing(container, expected):
    assert_frame_equal(container.data_testing, expected)


@pytest.mark.parametrize('container, args, kwargs, type_error, message_error', [
    # index flat
    (
        tahini.core.base.ContainerDataIndexedMultiSets(),
        [],
        dict(index=[0]),
        TypeError,
        "object of type 'int' has no len()",
    ),
    # index not exist
    (
        tahini.core.base.ContainerDataIndexedMultiSets(),
        [],
        dict(index=[(0, 1)]),
        KeyError,
        "[frozenset({0, 1})] not found in axis",
    ),
])
def test_container_data_indexed_multi_sets_drop_error(container, args, kwargs, type_error, message_error):
    with pytest.raises(type_error) as e:
        container.drop(*args, **kwargs)
    assert e.value.args[0] == message_error


@pytest.mark.parametrize('container, args, kwargs, expected', [
    # empty
    (
        tahini.core.base.ContainerDataIndexedMultiSets(),
        [],
        dict(),
        tahini.core.base.ContainerDataIndexedMultiSets(),
    ),
    # empty container ignore error
    (
        tahini.core.base.ContainerDataIndexedMultiSets(),
        [],
        dict(index=[(0, 1)], errors='ignore'),
        tahini.core.base.ContainerDataIndexedMultiSets(),
    ),
    # empty inputs
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        [],
        dict(),
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
    ),
    # non empty
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        [],
        dict(index=[(0, 1)]),
        tahini.core.base.ContainerDataIndexedMultiSets(),
    ),
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1), (0, 2)]),
        [],
        dict(index=[(0, 1), (0, 2)]),
        tahini.core.base.ContainerDataIndexedMultiSets(),
    ),
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1), (0, 2)]),
        [],
        dict(index=[(0, 2)]),
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
    ),
    # order does not matter
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        [],
        dict(index=[(1, 0)]),
        tahini.core.base.ContainerDataIndexedMultiSets(),
    ),
    # drop columns
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)], columns=['a']),
        [],
        dict(columns=['a']),
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
    ),
])
def test_container_data_indexed_multi_sets_drop(container, args, kwargs, expected):
    container = container.drop(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(container, expected)


@pytest.mark.parametrize('container, args, kwargs, expected', [
    # empty
    (tahini.core.base.ContainerDataIndexedMultiSets(), [], dict(), tahini.core.base.ContainerDataIndexedMultiSets()),
    # empty inputs
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        [],
        dict(),
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
    ),
    # todo fix this test
    # (
    #     tahini.container.ContainerDataIndexedMultiSets(index=[(0, 1)]),
    #     [],
    #     dict(index=[(0, 2)]),
    #     tahini.container.ContainerDataIndexedMultiSets(index=[(0, 1), (0, 2)]),
    # ),
    # non empty
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        [],
        dict(index=[(0, 2)]),
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 2), (0, 1)]),
    ),
    # order does not matter
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
        [],
        dict(index=[(1, 0)]),
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1)]),
    ),
])
def test_container_data_indexed_multi_sets_update(container, args, kwargs, expected):
    container = container.update(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(container, expected)


@pytest.mark.parametrize('container, args, kwargs, expected', [
    # empty
    (
        tahini.core.base.ContainerDataIndexedMultiSets(),
        [],
        dict(),
        tahini.core.base.ContainerDataIndexedMultiSets(),
    ),
    # non empty
    (
        tahini.core.base.ContainerDataIndexedMultiSets(index=[(0, 1), (0, 2)]),
        [],
        dict(mapper={0: 'a', 1: 'b', 2: 'c'}),
        tahini.core.base.ContainerDataIndexedMultiSets(index=[('a', 'b'), ('a', 'c')]),
    ),
])
def test_container_data_indexed_multi_sets_map(container, args, kwargs, expected):
    container = container.map(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(container, expected)



from functools import partial

import pytest
import pandas as pd
import plotly.graph_objects as go

import tahini.plot.engine
import tahini.core.graph
import tahini.testing

name_nodes = tahini.core.graph.Graph().nodes.names_index[0]

assert_frame_equal = partial(
    pd.testing.assert_frame_equal,
    check_dtype=False,
    check_column_type=False,
    check_index_type=False,
)


@pytest.mark.parametrize('args, kwargs, expected', [
    # empty graph
    ([], dict(graph=tahini.core.graph.Graph()), tahini.core.graph.Graph())
])
def test_base_init(args, kwargs, expected):
    engine = tahini.plot.engine.Base(*args, **kwargs)
    tahini.testing.assert_graph_equal(engine.graph, expected)


@pytest.mark.parametrize('engine, args, kwargs, expected', [
    # empty graph
    (
        tahini.plot.engine.Base(graph=tahini.core.graph.Graph()),
        [],
        dict(),
        pd.DataFrame(columns=['position_dim_0', 'position_dim_1'], index=pd.Index([], name=name_nodes)),
    ),
    # columns in graph
    (
        tahini.plot.engine.Base(graph=tahini.core.graph.Graph(nodes_data=dict(a=[]))),
        [],
        dict(),
        pd.DataFrame(columns=['a', 'position_dim_0', 'position_dim_1'], index=pd.Index([], name=name_nodes)),
    ),
])
def test_base_get_data_frame_plot(engine, args, kwargs, expected):
    df = engine.get_data_frame_plot(*args, **kwargs)
    assert_frame_equal(df, expected)


@pytest.mark.parametrize('engine, args, kwargs', [
    # non empty graph
    (tahini.plot.engine.Plotly(graph=tahini.core.graph.Graph(order=10)), [], dict()),
    # empty
    (tahini.plot.engine.Plotly(graph=tahini.core.graph.Graph()), [], dict()),
])
def test_plot(engine, args, kwargs):
    fig = engine.plot(*args, **kwargs)
    assert isinstance(fig, go.Figure)



import pytest

import tahini.core.edges
import tahini.core.nodes
import tahini.testing


@pytest.mark.parametrize('args, kwargs', [
    # empty
    ([], dict()),
    # non empty
    ([], dict(index=[(0, 1)])),
])
def test_edges_init_simple(args, kwargs):
    tahini.core.edges.Edges(*args, **kwargs)


@pytest.mark.parametrize('args, kwargs', [
    # empty
    ([], dict()),
    # non empty
    ([], dict(index=[(0, 1)])),
])
def test_undirected_edges_init_simple(args, kwargs):
    tahini.core.edges.UndirectedEdges(*args, **kwargs)


@pytest.mark.parametrize('edges, args, kwargs, expected', [
    # empty
    (tahini.core.edges.Edges(), [], dict(), tahini.core.nodes.Nodes()),
    (tahini.core.edges.UndirectedEdges(), [], dict(), tahini.core.nodes.Nodes()),
    # non empty
    (tahini.core.edges.Edges(index=[(0, 1)]), [], dict(), tahini.core.nodes.Nodes(index=[0, 1])),
    (tahini.core.edges.Edges(index=[(0, 1), (0, 2)]), [], dict(), tahini.core.nodes.Nodes(index=[0, 1, 2])),
    (tahini.core.edges.UndirectedEdges(index=[(0, 1)]), [], dict(), tahini.core.nodes.Nodes(index=[0, 1])),
    (tahini.core.edges.UndirectedEdges(index=[(0, 1), (0, 2)]), [], dict(), tahini.core.nodes.Nodes(index=[0, 1, 2])),
    # order matters
    (tahini.core.edges.Edges(index=[(0, 2), (0, 1)]), [], dict(), tahini.core.nodes.Nodes(index=[0, 2, 1])),
])
def test_edges_nodes(edges, args, kwargs, expected):
    nodes = edges.nodes
    tahini.testing.testing.assert_container_equal(nodes, expected)


@pytest.mark.parametrize('edges, args, kwargs, expected', [
    # empty
    (tahini.core.edges.Edges(), [], dict(), tahini.core.edges.Edges()),
    (tahini.core.edges.UndirectedEdges(), [], dict(), tahini.core.edges.UndirectedEdges()),
    # non empty inputs
    (tahini.core.edges.Edges(), [], dict(nodes=[0]), tahini.core.edges.Edges()),
    (tahini.core.edges.UndirectedEdges(), [], dict(nodes=[0]), tahini.core.edges.UndirectedEdges()),
    # non empty
    (tahini.core.edges.Edges(index=[(0, 1)]), [], dict(nodes=[0]), tahini.core.edges.Edges()),
    (tahini.core.edges.UndirectedEdges(index=[(0, 1)]), [], dict(nodes=[0]), tahini.core.edges.UndirectedEdges()),
    (tahini.core.edges.Edges(index=[(0, 1)]), [], dict(nodes=[0, 1]), tahini.core.edges.Edges(index=[(0, 1)])),
    (
        tahini.core.edges.UndirectedEdges(index=[(0, 1)]),
        [],
        dict(nodes=[0, 1]),
        tahini.core.edges.UndirectedEdges(index=[(0, 1)]),
    ),
    (tahini.core.edges.Edges(index=[(0, 1), (0, 2)]), [], dict(nodes=[0, 1]), tahini.core.edges.Edges(index=[(0, 1)])),
    (
        tahini.core.edges.UndirectedEdges(index=[(0, 1), (0, 2)]),
        [],
        dict(nodes=[0, 1]),
        tahini.core.edges.UndirectedEdges(index=[(0, 1)]),
    ),
])
def test_edges_keep_nodes(edges, args, kwargs, expected):
    edges_keep_nodes = edges.keep_nodes(*args, **kwargs)
    tahini.testing.testing.assert_container_equal(edges_keep_nodes, expected)



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

install_requires = [
    'pandas',
    'numpy',
]

extras_require = {
    'plotly': ['plotly'],
}

extras_require['all'] = sorted(set(sum(extras_require.values(), [])))

setup(
    name='tahini',
    version='2021.02.2',
    author='tahini-dev',
    author_email='tahini.dev@gmail.com',
    description='Python package for graph theory',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/tahini-dev/tahini',
    packages=find_packages(exclude=('tests',)),
    install_requires=install_requires,
    extras_require=extras_require,
    python_requires='>=3.7',
    include_package_data=True,
)
